/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 ************************************************************************
 *  \file
 *     fgs_decoder.h
 *  \brief
 *     definitions for FGS decoding
 *  \author
 *      - Ascenso, Joo                             <joao.ascenso@lx.it.pt>
 *
 * ************************************************************************
 */

#ifndef FGS_DECODER_H
#define FGS_DECODER_H

/**************************************************************************
Definitions for the DCT Transform
***************************************************************************/

#define CONST_BITS  13
#define PASS1_BITS  2
#define FIX_0_298631336  ((int)  2446)
#define FIX_0_390180644  ((int)  3196)
#define FIX_0_541196100  ((int)  4433)
#define FIX_0_765366865  ((int)  6270)
#define FIX_0_899976223  ((int)  7373)
#define FIX_1_175875602  ((int)  9633)
#define FIX_1_501321110  ((int)  12299)
#define FIX_1_847759065  ((int)  15137)
#define FIX_1_961570560  ((int)  16069)
#define FIX_2_053119869  ((int)  16819)
#define FIX_2_562915447  ((int)  20995)
#define FIX_3_072711026  ((int)  25172)

#define MULTIPLY(var,const)  ((var) * (const))
#define RIGHT_SHIFT(x,shft)	((x) >> (shft))
#define DCTSIZE		    8
#define DCTSIZE2	    64
#define MBSIZE		    16
#define BSIZE		      8
#define TSIZE		      4
#define TSIZE2	      16
#define ONE	((int) 1)
#define DESCALE(x,n)  RIGHT_SHIFT((x) + (ONE << ((n)-1)), n)

/**************************************************************************
FGS definitions
***************************************************************************/

#define FGS_VOP_START_CODE 0x01B9
#define FGS_VOP_BP_MARKER_HEADER_LENGTH 27
#define FGS_VOP_BP_MARKER_ID_LENGTH 5
#define FGS_VOP_BP_MARKER_HEADER 0x0A
#define VISUAL_OBJECT_SEQUENCE_END_CODE 0x01B1
#define VISUAL_OBJECT_SEQUENCE_END_CODE_LENGHT 32
#define MAX_BP_LEVEL 15                 //maximum number of bitplanes supported
#define ABS(x)        (((x) < 0) ? -(x) : (x))

typedef struct //cannot be called bitstream => there is already one
{
  int             byte_pos;           //!< current position in bitstream;
  int             bit_pos;            //!< current bitcounter
  int             byte_pos_book;      //!< bookmark of byte position in the bitstream
  int             bit_pos_book;       //!< bookmark of bit position in the bitstream
  byte            *streamBuffer;      //!< actual buffer for read bytes
  int             read_bytes;         //!< number of read bytes from file (relative to a frame buffer)
  int             p_frame;            //!< pointer to file to where the last frame occurred
  int             buffer_size;        //!< size of the buffer
  FILE            *pFGSFile;          //!< pointer to bitstream file
}decBitstream;

// to mask the n least significant bits of a byte or integer
static unsigned int mask_bits[9] =
{
    0x00000001, 0x00000003, 0x00000007,
    0x0000000f, 0x0000001f, 0x0000003f, 0x0000007f,
    0x000000ff
};

typedef struct enh_block
{
  int run[4][TSIZE2];                 //!< number of runs for each level
  int end[4][TSIZE2];                 //!< end = 0 or 1 for each level
  int sign[4][TSIZE2];                //!< sign of the coefficient in the block
  int residual_sign[4][TSIZE2];       //!< temporary array to see if the sign was already read
  int data_level;                     //!< number of bitplanes necessary to represent the block
  byte cbp;                           //!< code bit pattern for each block
  int decoded_level;                  //!< decoded level
} enh_block;

//FGS data struct
typedef struct
{
  byte     maximum_level_y_ac;        //!< maximum number of bitplanes of the Y component 
  byte     maximum_level_y_dc;        //!< maximum number of bitplanes of the Y component 

  byte     maximum_level_chr_dc;        //!< maximum number of bitplanes of the U component
  byte     maximum_level_chr_ac;        //!< maximum number of bitplanes of the V component

  byte     NumberOfBitplanes;         //!< maximum number of bitplanes coded
  byte     maximum_level;             //!< maximum number of bitplanes of all components

  enh_block **block_data_lum_ac;      //!< pointer to vector of enhancenment blocks
  enh_block **block_data_u_ac;
  enh_block **block_data_v_ac;

  enh_block **block_data_lum_dc;      //!< pointer to vector of enhancenment blocks
  enh_block **block_data_u_dc;
  enh_block **block_data_v_dc;


  int hDCTMBlocks;              //!< number of lum Mblocks in height
  int wDCTMBlocks;              //!< number of lum Mblocks in width

  int hDCTBlocks_cr;                  //!< number of chr blocks in height
  int wDCTBlocks_cr;                  //!< number of chr blocks in width

  byte *except_code;                  //!< exception code for fgs_msb_not_reached 

  int      **Ydiff_ac;                //!< residual image Y (AC components)
  int      **Udiff_ac;                //!< residual image U (AC components)
  int      **Vdiff_ac;                //!< residual image V (AC components)

  int      **Ydiff_dc;                //!< residual image Y (AC components)
  int      *Udiff_dc;                 //!< residual image U (AC components)
  int      *Vdiff_dc;                 //!< residual image V (AC components)

  byte **YBitplane_ac;            //!< Y,U and V bitplane (AC components)
  byte **UBitplane_ac;
  byte **VBitplane_ac;
  
  byte **YBitplane_dc;            //!< Y,U and V bitplane (DC components)
  byte *UBitplane_dc;
  byte *VBitplane_dc;

  byte     **YSignal_ac;                 //!< signal (+-) of the Y component
  byte     **USignal_ac;                 //!< signal (+-) of the U component
  byte     **VSignal_ac;                 //!< signal (+-) of the V component

  byte     **YSignal_dc;                 //!< signal (+-) of the Y component
  byte     *USignal_dc;                 //!< signal (+-) of the U component
  byte     *VSignal_dc;                 //!< signal (+-) of the V component

  Boolean skip_y_ac;                 //!< flag to indicate encoding of the y component
  Boolean skip_y_dc;                 //!< flag to indicate encoding of the y component

  Boolean skip_chr_dc;                 //!< flag to indicate encoding of the u component
  Boolean skip_chr_ac;                 //!< flag to indicate encoding of the v component
  
  int endOfData;

}FGSStruct;

/**************************************************************************
Prototype definitions           
***************************************************************************/

void initFGS (struct inp_par *inp,struct img_par *pimg);
void decodeFGSFrame (struct img_par *pimg);
void decodeFGSFrameHeader();
int decodeFGSData(struct img_par *pimg);
int decodeBitplane(struct img_par *pimg,int level,int max_level);
void calcBitplane(struct img_par *pimg,int level);
void finishEncoding(struct img_par *pimg);
void addResidueToImg(struct img_par *pimg);
void copyFGSframe(struct img_par *img);
void writePrevFGSFrame(struct img_par *img);
void writeFGSFrame(struct img_par *img);

//inverse transform related functions
void calcitrans_lum(int ** dataAC, int ** dataDC,int hDCTBlocks,int wDCTBlocks);
void calcitrans_chr(int ** dataAC, int * dataDC,int hDCTBlocks,int wDCTBlocks);
void H26L_itransChr(int **coefAC,int *coefDC);
void H26L_itransLum(int **coefAC,int **coefDC);
void zigzag_scan(int hIndex,int wIndex,int i,byte **buffer,byte **bitplane);
void scanSignal(int hIndex,int wIndex,int sign[][TSIZE2],byte **bitplane);
void recoverSignal(int hIndex,int wIndex,int value,int **bitplane);
void idctInteger (int **data);
void cleanFGS();
void clearData(struct img_par *img);
void endCalcBitplane(struct img_par *pimg,int nBp);
void rleDecoding(enh_block* block_data,byte *out,int t_block);
void rleDecoding2x2(enh_block* block_data,byte *out,int t_block);
int getMBIndex(int level);
int readCbpindexMSB(int comb_index,int block_index,byte *except_code);
int readCbpindex(int comb_index,int block_index,byte *except_code);
int decodeBlockBP(enh_block* block_data, int level,int max_level);
int decodeBlockBP_chrDC(enh_block* block_data, int level, int max_level);
int decodeBlockBP_lumDC(enh_block* block_data, int level);
int decodeVLCCode(int *value);

// Bitstream buffer management functions
int readByteFGS(byte *data,unsigned int numBits, decBitstream *pBitstream);
int readIntFGS(unsigned int *data, int nbits, decBitstream *pBitstream);
void peekIntFGS(unsigned int *data, int numBits, decBitstream *pBitstream);
void peekByteFGS(byte *data,unsigned int numBits, decBitstream *pBitstream);
void readStuffingBits(decBitstream *pBitstream,int check);
void initializeBitstream(struct inp_par *inp);
void closeBitstream();
void addBookmark(decBitstream *pBitstream);
void getBookmark(decBitstream *pBitstream);
void reallocBuffer(decBitstream *pBitstream);
void readNextFrame(decBitstream *pBitstream, int pdata);
int checkEnd(decBitstream *pBitstream);

//debug !!!!
void debug_residue(ImageParameters *pimg);

#endif