/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 ***********************************************************************
 *  \mainpage
 *     This is the H.26L decoder reference software. For detailed documentation
 *     see the comments in each file.
 *
 *  \author
 *     The main contributors are listed in contributors.h
 *
 *  \version
 *     JM 1.6
 *
 *  \note
 *     tags are used for document system "doxygen"
 *     available at http://www.doxygen.org
 */
/*!
 *  \file
 *     fgs_decoder.c
 *  \brief
 *     blah
 *  \author
 *   - Joo Ascenso               <joao.ascenso@lx.it.pt>
 *
 ***********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

#include "global.h"
#include "memalloc.h"
#include "fgs_decoder.h"

//Trace FILE
//#define FGS_TRACE
#ifdef FGS_TRACE
int gtr;
#define FGS_TRACE_PATH "d:\\traceFGSD.txt"
FILE *pFGSTraceFile;
#endif
//Data
FGSStruct *pFGSData;
decBitstream *pFGSBitstream;

//array to indicate the zigzag scan order
const int fgs_zigzag[16] = {    //array to indicate the zigzag scan order
  0, 1, 5, 6,
  2, 4, 7,12,
  3, 8,11,13,
  9,10,14,15,
};

static const int fgs_dequant_coef[4][4] = {
  {16,20,16,20},
  {20,26,20,26},
  {16,20,16,20},
  {20,26,20,26}
};

/**************************************************************************
ULVC Tables to decode the cbp index and the (run,eop) codes
***************************************************************************/

const unsigned int table_cbp_yyyy[3][17] = {
	{0xFF,0x00,0x08,0x04,0x01,0x02,0x0C,0x03,0x0A,0x05,0x06,0x09,0x0E,0x07,0x0D,0x0B,0x0F},

  {0xFF,0x00,0x08,0x01,0x04,0x02,0x0C,0x03,0x05,0x0A,0x0F,0x06,0x07,0x09,0x0B,0x0D,0x0E},

  {0xFF,0x00,0x01,0x02,0x04,0x08,0x03,0x0C,0x05,0x0A,0x09,0x06,0x0D,0x0E,0x0B,0x07,0x0F},
};

const unsigned int table_cbp_yyyyuv[3][49] = {
  {0xFF,0x0,0x020,0x01,0x02,0x04,0x08,0x0C,0x03,0x0A,0x05,0x022,0x021,0x028,0x024,0x09,0x06,0x010,0x0E,
    0x0B,0x07,0x0D,0x0F,0x02A,0x02C,0x023,0x025,0x026,0x029,0x02E,0x02B,0x02F,0x02D,0x027,0x011,0x014,0x012,
    0x018,0x013,0x015,0x01A,0x01C,0x01B,0x016,0x019,0x01E,0x01D,0x01F,0x017},

  {0xFF,0x0,0x02F,0x0F,0x08,0x01,0x04,0x02,0x0C,0x03,0x05,0x010,0x0A,0x020,0x02B,0x02E,0x02D,0x0E,0x027,0x01F,
    0x07,0x0B,0x0D,0x02C,0x09,0x02A,0x06,0x025,0x023,0x022,0x028,0x024,0x021,0x029,0x026,0x01E,0x017,0x01B,
    0x01D,0x012,0x018,0x01C,0x014,0x013,0x01A,0x011,0x015,0x019,0x016},

  {0xFF,0x0,0x02F,0x0F,0x01,0x08,0x04,0x02,0x020,0x0C,0x02E,0x02D,0x02B,0x03,0x027,0x05,0x0A,0x01F,0x0E,0x07,
    0x0B,0x0D,0x02C,0x02A,0x025,0x023,0x09,0x06,0x022,0x028,0x021,0x024,0x010,0x029,0x026,0x01D,0x01E,0x01B,
    0x017,0x013,0x01C,0x01A,0x015,0x011,0x014,0x018,0x012,0x016,0x019},
};

const unsigned int table_dc_lum_uvlc[3][33] = {
  {0xFF,0x0F,0x010,0x011,0x012,0x013,0x015,0x00,0x014,0x019,0x016,0x018,0x017,0x01A,0x01,0x01C,0x01B,0x02,
    0x01E,0x01D,0x01F,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E},

  {0xFF,0x0F,0x00,0x010,0x01,0x02,0x011,0x012,0x03,0x013,0x014,0x015,0x04,0x016,0x05,0x017,0x019,0x018,0x06,
    0x01A,0x07,0x01B,0x08,0x01C,0x09,0x01D,0x01E,0x0A,0x01F,0x0B,0x0C,0x0D,0x0E},

  {0xFF,0x00,0x01,0x02,0x010,0x03,0x011,0x04,0x012,0x05,0x013,0x014,0x06,0x015,0x0F,0x07,0x016,0x08,0x017,0x018,
    0x09,0x019,0x01A,0x0A,0x01B,0x0B,0x01C,0x01D,0x0C,0x01E,0x0D,0x01F,0x0E},
};

const unsigned int table_dc_chr_uvlc[3][9] = {
  {0xFF,0x04,0x00,0x03,0x05,0x06,0x01,0x07,0x02},

  {0xFF,0x00,0x04,0x03,0x05,0x01,0x06,0x07,0x02},

  {0xFF,0x00,0x04,0x05,0x01,0x06,0x03,0x07,0x02},
};

const unsigned int table_ac_uvlc[3][33] = {
  {0xFF,0x0F,0x01,0x012,0x011,0x02,0x00,0x010,0x013,0x015,0x03,0x014,0x04,0x05,0x016,0x017,0x018,0x019,0x06,
    0x07,0x01B,0x08,0x01A,0x09,0x01C,0x01E,0x0A,0x01D,0x0B,0x0C,0x01F,0x0D,0x0E},

  {0xFF,0x01,0x00,0x02,0x010,0x03,0x011,0x0F,0x012,0x04,0x013,0x014,0x05,0x015,0x016,0x06,0x017,0x07,0x018,
    0x019,0x08,0x09,0x01A,0x01B,0x0A,0x01C,0x0B,0x01E,0x01D,0x0C,0x01F,0x0D,0x0E},

  {0xFF,0x00,0x01,0x02,0x010,0x03,0x011,0x04,0x012,0x013,0x05,0x014,0x06,0x015,0x0F,0x016,0x07,0x017,0x08,
    0x018,0x019,0x09,0x01A,0x0A,0x01B,0x0B,0x01C,0x0C,0x01E,0x01D,0x0D,0x01F,0x0E},
};

void initFGS (struct inp_par *inp,struct img_par *pimg)
{
  int a,nblocks;

  printf("FGS decoder initialization\n");
  //alloc FGS data struct
  pFGSData = (FGSStruct *) calloc(1, sizeof(FGSStruct));
  //alloc residual images
  get_mem2Dint(&pFGSData->Ydiff_ac, pimg->height, pimg->width);
  get_mem2Dint(&pFGSData->Udiff_ac, pimg->height_cr, pimg->width_cr);
  get_mem2Dint(&pFGSData->Vdiff_ac, pimg->height_cr, pimg->width_cr);

  get_mem2Dint(&pFGSData->Ydiff_dc, pimg->height>>2, pimg->width>>2);
  pFGSData->Udiff_dc = (int *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(int));
  pFGSData->Vdiff_dc = (int *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(int));
  //alloc biplanes
  get_mem2D(&pFGSData->YBitplane_ac, pimg->height, pimg->width);
  get_mem2D(&pFGSData->UBitplane_ac, pimg->height_cr, pimg->width_cr);
  get_mem2D(&pFGSData->VBitplane_ac, pimg->height_cr, pimg->width_cr);

  get_mem2D(&pFGSData->YBitplane_dc, pimg->height>>2, pimg->width>>2);
  pFGSData->UBitplane_dc = (byte *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(byte));
  pFGSData->VBitplane_dc = (byte *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(byte));
  //alloc signal
  get_mem2D(&pFGSData->YSignal_ac, pimg->height, pimg->width);
  get_mem2D(&pFGSData->USignal_ac, pimg->height_cr, pimg->width_cr);
  get_mem2D(&pFGSData->VSignal_ac, pimg->height_cr, pimg->width_cr);

  get_mem2D(&pFGSData->YSignal_dc, pimg->height>>2, pimg->width>>2);
  pFGSData->USignal_dc = (byte *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(byte));
  pFGSData->VSignal_dc = (byte *) calloc((pimg->height_cr>>2)*(pimg->width_cr>>2), sizeof(byte));
  //initialize bitstream buffer
  initializeBitstream(inp);
  
  pFGSData->hDCTMBlocks = pimg->height/MBSIZE;
  pFGSData->wDCTMBlocks = pimg->width/MBSIZE;

  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;

  //allocate structure with encoding data for luminance
  pFGSData->block_data_lum_ac = (enh_block **)calloc(4*nblocks,sizeof(enh_block *));
  if(pFGSData->block_data_lum_ac == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < 4*nblocks;a++)
  {
    pFGSData->block_data_lum_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_lum_ac[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  pFGSData->block_data_lum_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  if(pFGSData->block_data_lum_dc == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < nblocks;a++)
  {
    pFGSData->block_data_lum_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_lum_dc[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  pFGSData->hDCTBlocks_cr = pimg->height_cr/BSIZE;
  pFGSData->wDCTBlocks_cr = pimg->width_cr/BSIZE;

  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;

  //allocate structure with encoding data for chrominance components
  pFGSData->block_data_u_ac = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  pFGSData->block_data_v_ac = (enh_block **)calloc(nblocks,sizeof(enh_block *));

  if(pFGSData->block_data_u_ac == NULL || pFGSData->block_data_v_ac == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  pFGSData->block_data_u_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  pFGSData->block_data_v_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));

  if(pFGSData->block_data_u_dc == NULL || pFGSData->block_data_v_dc == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < nblocks;a++)
  {
    pFGSData->block_data_u_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    pFGSData->block_data_v_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_u_ac[a] == NULL || pFGSData->block_data_v_ac[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");

    pFGSData->block_data_u_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    pFGSData->block_data_v_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_u_dc[a] == NULL || pFGSData->block_data_v_dc[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  //alloc exception code ... total number of macroblocks float
  pFGSData->except_code = (byte *) calloc((int)ceil(pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks), sizeof(byte));
  if(pFGSData->except_code == NULL)
    no_mem_exit("Memory allocation failure: initFGS"); 
}


void decodeFGSFrame (struct img_par *pimg)
{
  int hDCTBlocks,wDCTBlocks,end;

  decodeFGSFrameHeader();
  end = decodeFGSData(pimg);

  //because the stuffing bits in truncated bitstreams are wrong... 
  if(end == -1) readStuffingBits(pFGSBitstream,1);
  else readStuffingBits(pFGSBitstream,0);

  if(checkEnd(pFGSBitstream) != -1) pFGSBitstream->byte_pos++; //ugly hack !!! 
  if(checkEnd(pFGSBitstream) != -1) pFGSBitstream->byte_pos-=2; //ugly hack !!! 
  if(checkEnd(pFGSBitstream) != -1) pFGSBitstream->byte_pos--; //ugly hack !!! 
  if(checkEnd(pFGSBitstream) != -1) pFGSBitstream->byte_pos--; //ugly hack !!! 
  if(checkEnd(pFGSBitstream) != -1) printf("Can't synchronize\n");
  
  readNextFrame(pFGSBitstream,pFGSBitstream->byte_pos);

  hDCTBlocks = pimg->height/MBSIZE;
  wDCTBlocks = pimg->width/MBSIZE;
 
  //Calculate the IDCT on the Y image

  calcitrans_lum(pFGSData->Ydiff_ac,pFGSData->Ydiff_dc,hDCTBlocks,wDCTBlocks);
 
  hDCTBlocks = pimg->height_cr/BSIZE;
  wDCTBlocks = pimg->width_cr/BSIZE;

  //Calculate the IDCT on the U e V image
  calcitrans_chr(pFGSData->Udiff_ac,pFGSData->Udiff_dc,hDCTBlocks,wDCTBlocks);
  calcitrans_chr(pFGSData->Vdiff_ac,pFGSData->Vdiff_dc,hDCTBlocks,wDCTBlocks);
}

void decodeFGSFrameHeader()
{
  int start_code;
  byte temp;
  pFGSData->endOfData = 1;
  //read start code
  readIntFGS(&start_code,32,pFGSBitstream);
  if(start_code != FGS_VOP_START_CODE)
  {
    snprintf(errortext, ET_SIZE, "decodeFGSFrameHeader: Wrong Start Code\n");
    error(errortext, 100);
  }
  //read type of frames
  readByteFGS(&temp,2,pFGSBitstream);
  //read temporal reference
  readByteFGS(&temp,8,pFGSBitstream);
  fprintf(stdout,"FGS Temporal Reference: %d\n",temp);
#ifdef FGS_TRACE
  gtr = temp;
  fprintf(pFGSTraceFile,"Frame number %d\n",temp);
  fflush(pFGSTraceFile);
#endif
  //marker bit
  readByteFGS(&temp,1,pFGSBitstream);
  //read fgs_vop_max_level_y,u and v
  readByteFGS(&pFGSData->maximum_level_y_dc,5,pFGSBitstream);
  readByteFGS(&pFGSData->maximum_level_y_ac,5,pFGSBitstream);

  readByteFGS(&pFGSData->maximum_level_chr_dc,5,pFGSBitstream);
  readByteFGS(&pFGSData->maximum_level_chr_ac,5,pFGSBitstream);

  //marker bit
  readByteFGS(&temp,1,pFGSBitstream);
  //read fgs_vop_number_of_vop_bp_coded
  readByteFGS(&pFGSData->NumberOfBitplanes,5,pFGSBitstream);
  //read fgs_vop_mc_bit_plane_used.. => not used
/*readByteFGS(&temp,5,pFGSBitstream);
  //fgs_vop_selective_enhancement.. => not used
  readByteFGS(&temp,1,pFGSBitstream);*/
  readStuffingBits(pFGSBitstream,0);
  if(pFGSData->endOfData == -1)
  {
    snprintf(errortext, ET_SIZE, "decodeFGSFrameHeader: Insufficient Data\n");
    error(errortext, 100);
  }
}

int decodeFGSData(struct img_par *pimg)
{
  int max_level,temp,i,j,nblocks,end,max_level_y,max_level_chr;

  max_level_y = max(pFGSData->maximum_level_y_ac,pFGSData->maximum_level_y_dc);
  max_level_chr = max(pFGSData->maximum_level_chr_ac,pFGSData->maximum_level_chr_dc);
  pFGSData->maximum_level = max(max_level_y,max_level_chr);

  if (pFGSData->NumberOfBitplanes > pFGSData->maximum_level)
        max_level = pFGSData->maximum_level;
  else
        max_level = pFGSData->NumberOfBitplanes;

  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;
  memset(pFGSData->except_code,0,nblocks*sizeof(byte));

  printf("DC level:%d AC Level: %d\n",pFGSData->maximum_level_y_dc,
    max(pFGSData->maximum_level_y_ac,max(pFGSData->maximum_level_chr_ac,pFGSData->maximum_level_chr_dc)));

  //data level is initialized to -1 and cbp to 1 => zero blocks present
  for ( i=0; i < 4*nblocks; i++ ) 
  {
    pFGSData->block_data_lum_ac[i]->data_level = -1;
    pFGSData->block_data_lum_ac[i]->cbp = 0;
    for(j=0;j<4;j++)
    {
      memset(pFGSData->block_data_lum_ac[i]->sign[j],0,TSIZE2*sizeof(int));
      memset(pFGSData->block_data_lum_ac[i]->residual_sign[j],0,TSIZE2*sizeof(int));
    }
  }
  for ( i=0; i < nblocks; i++ ) 
  {  
    pFGSData->block_data_lum_dc[i]->cbp = 0;
    pFGSData->block_data_lum_dc[i]->data_level = -1;
    memset(pFGSData->block_data_lum_dc[i]->sign[0],0,TSIZE2*sizeof(int));
    memset(pFGSData->block_data_lum_dc[i]->residual_sign[0],0,TSIZE2*sizeof(int));
  }

  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;
  for (i=0; i < nblocks; i++ )
  {
    pFGSData->block_data_u_ac[i]->data_level = -1;
    pFGSData->block_data_v_ac[i]->data_level = -1;
    pFGSData->block_data_u_dc[i]->data_level = -1;
    pFGSData->block_data_v_dc[i]->data_level = -1;
    pFGSData->block_data_u_ac[i]->cbp = 0;
    pFGSData->block_data_v_ac[i]->cbp = 0;
    pFGSData->block_data_u_dc[i]->cbp = 0;
    pFGSData->block_data_v_dc[i]->cbp = 0;
    for(j=0;j<4;j++)
    {
      memset(pFGSData->block_data_u_ac[i]->sign[j],0,TSIZE2*sizeof(int));
      memset(pFGSData->block_data_u_ac[i]->residual_sign[j],0,TSIZE2*sizeof(int));
      memset(pFGSData->block_data_v_ac[i]->sign[j],0,TSIZE2*sizeof(int));
      memset(pFGSData->block_data_v_ac[i]->residual_sign[j],0,TSIZE2*sizeof(int));
    }
    memset(pFGSData->block_data_u_dc[i]->sign[0],0,TSIZE2*sizeof(int));
    memset(pFGSData->block_data_u_dc[i]->residual_sign[0],0,TSIZE2*sizeof(int));
    memset(pFGSData->block_data_v_dc[i]->sign[0],0,TSIZE2*sizeof(int));
    memset(pFGSData->block_data_v_dc[i]->residual_sign[0],0,TSIZE2*sizeof(int));
  }
  
  //put the decoded level at maximum => complete decoding of the bitstream

  for(i=0; i<pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks; i++)
  {
    pFGSData->block_data_lum_dc[i/4]->decoded_level = max_level;
    pFGSData->block_data_lum_ac[i]->decoded_level = max_level;
    pFGSData->block_data_lum_ac[i+1]->decoded_level = max_level;
    pFGSData->block_data_lum_ac[i+2]->decoded_level = max_level;
    pFGSData->block_data_lum_ac[i+3]->decoded_level = max_level;
    pFGSData->block_data_u_dc[i/4]->decoded_level = max_level;
    pFGSData->block_data_v_dc[i/4]->decoded_level = max_level;
    pFGSData->block_data_u_ac[i/4]->decoded_level = max_level;
    pFGSData->block_data_v_ac[i/4]->decoded_level = max_level;
  }
  clearData(pimg);
  end = 1;
  for (i=0; i<max_level; i++)
  {
    //Bitplane start code
    end = readIntFGS(&temp,FGS_VOP_BP_MARKER_HEADER_LENGTH,pFGSBitstream);
    if(end == -1) break;
    if(temp != FGS_VOP_BP_MARKER_HEADER)
    {
      printf("FGS_VOP_BP_MARKER_HEADER not present: decodeFGSData, level=%d\n",i);
      exit(1);
    }
    if(readIntFGS(&temp,FGS_VOP_BP_MARKER_ID_LENGTH,pFGSBitstream) == -1)
    {
      end = -1;
      break;
    }
    end = decodeBitplane(pimg,i,max_level);
    calcBitplane(pimg,i);
    if(end == -1) break;
  }
  endCalcBitplane(pimg,max_level-(i+1));
  finishEncoding(pimg);
  return end;
}

int decodeBitplane(struct img_par *pimg,int level,int max_level)
{
  int end,a,b,c,d,i,j,mbindex,level_shift;
  byte **yBuffer_ac,**uBuffer_ac,**vBuffer_ac;
  byte *yBuffer_dc,*uBuffer_dc,*vBuffer_dc;

  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **ldata_dc = pFGSData->block_data_lum_dc;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

   //allocate temporary buffers
  get_mem2D(&yBuffer_ac,16,TSIZE2);
  get_mem2D(&uBuffer_ac,4,TSIZE2);
  get_mem2D(&vBuffer_ac,4,TSIZE2);
  
  yBuffer_dc = (byte *) calloc(TSIZE2, sizeof(byte));
  uBuffer_dc = (byte *) calloc(4, sizeof(byte));
  vBuffer_dc = (byte *) calloc(4, sizeof(byte));

  //see which components are skipped...
  level_shift = pFGSData->maximum_level-1-level;
  pFGSData->skip_y_ac = TRUE;
  pFGSData->skip_y_dc = TRUE;
  pFGSData->skip_chr_dc = TRUE;
  pFGSData->skip_chr_ac = TRUE;

  if((pFGSData->maximum_level_y_ac-1) >= level_shift) pFGSData->skip_y_ac = FALSE;
  if((pFGSData->maximum_level_y_dc-1) >= level_shift) pFGSData->skip_y_dc = FALSE;

  if((pFGSData->maximum_level_chr_dc-1) >= level_shift) pFGSData->skip_chr_dc = FALSE;
  if((pFGSData->maximum_level_chr_ac-1) >= level_shift) pFGSData->skip_chr_ac = FALSE;

  mbindex = getMBIndex(pFGSData->maximum_level-level);

  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      //encoding of the DC values
#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSTraceFile,"Lum DC block %d at level %d\n",i,level);
        fflush(pFGSTraceFile);
      }
#endif
      if(pFGSData->skip_y_dc == FALSE)
      {        
        end = decodeBlockBP_lumDC(ldata_dc[i],level);
        if(end == -1) break;
        ldata_dc[i]->decoded_level = max_level - level;

        rleDecoding(ldata_dc[i],yBuffer_dc,0);
      }
      //scanning of the DC values
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          pFGSData->YBitplane_dc[a*TSIZE+c][b*TSIZE+d] = yBuffer_dc[fgs_zigzag[c*TSIZE+d]];
        }
      }
      i++;
    }
    if(end == -1) break;
  }
  //do not encode more (there is only DC luminance values in this bitplane)
  if(mbindex == -1 || end == -1) 
  {
    free_mem2D(yBuffer_ac);
    free_mem2D(uBuffer_ac);
    free_mem2D(vBuffer_ac);
    free(yBuffer_dc);
    free(uBuffer_dc);
    free(vBuffer_dc);  
    return end;  
  }

  i=0;end=1;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {      
      if(level == 0 || level == 1)  //only for the MSB and MSB-2
      {
        if(level == 0)
        {
          end = readCbpindexMSB(mbindex,i,pFGSData->except_code);
          if(end == -1) break;
        }
        else
        {
          end = readCbpindex(mbindex,i,pFGSData->except_code);
          if(end == -1) break;
        }
      }
      //entropy encoding
      if(pFGSData->skip_y_ac == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Lum AC block %d at level %d\n",i,level);
          fflush(pFGSTraceFile);
        }
#endif
        end = decodeBlockBP(ldata_ac[i],level,max_level);
        if(end == -1) break;
        ldata_ac[i]->decoded_level = max_level - level;
        
        for(j=0;j < 4; j++)
          rleDecoding(ldata_ac[i],yBuffer_ac[j],j);
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Lum AC block %d at level %d\n",i+1,level);
          fflush(pFGSTraceFile);
        }
#endif
        end = decodeBlockBP(ldata_ac[i+1],level,max_level);
        if(end == -1) break;
        ldata_ac[i+1]->decoded_level = max_level - level;

        for(j=0;j < 4; j++)
          rleDecoding(ldata_ac[i+1],yBuffer_ac[4+j],j);
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Lum AC block %d at level %d\n",i+2,level);
          fflush(pFGSTraceFile);
        }
#endif
        end = decodeBlockBP(ldata_ac[i+2],level,max_level);
        if(end == -1) break;
        ldata_ac[i+2]->decoded_level = max_level - level;

        for(j=0;j < 4; j++)
          rleDecoding(ldata_ac[i+2],yBuffer_ac[8+j],j);
#ifdef FGS_TRACE        
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Lum AC block %d at level %d\n",i+3,level);
          fflush(pFGSTraceFile);
        }
#endif
        end = decodeBlockBP(ldata_ac[i+3],level,max_level);
        if(end == -1) break;
        ldata_ac[i+3]->decoded_level = max_level - level;

        for(j=0;j < 4; j++)
          rleDecoding(ldata_ac[i+3],yBuffer_ac[12+j],j);
      }
      if(pFGSData->skip_chr_dc == FALSE || pFGSData->skip_chr_ac == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Chr U DC block %d at level %d\n",i/4,level);
          fflush(pFGSTraceFile);    
        }
#endif
        end = decodeBlockBP_chrDC(udata_dc[i/4],level,max_level);
        if(end == -1) break;
        udata_dc[i/4]->decoded_level = max_level - level;

        rleDecoding2x2(udata_dc[i/4],uBuffer_dc,0);

#ifdef FGS_TRACE        
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Chr V DC block %d at level %d\n",i/4,level);
          fflush(pFGSTraceFile);
        }
#endif
        end = decodeBlockBP_chrDC(vdata_dc[i/4],level,max_level);
        if(end == -1) break;
        vdata_dc[i/4]->decoded_level = max_level - level;

        rleDecoding2x2(vdata_dc[i/4],vBuffer_dc,0);
      }
      if(pFGSData->skip_chr_ac == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Chr AC block %d at level %d\n",i/4,level);
          fflush(pFGSTraceFile); 
        }
#endif
        end = decodeBlockBP(udata_ac[i/4],level,max_level);
        if(end == -1) break;
        udata_ac[i/4]->decoded_level = max_level - level;
        
        for(j=0;j < 4; j++)
          rleDecoding(udata_ac[i/4],uBuffer_ac[j],j);
      
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Chr AC block %d at level %d\n",i/4,level);
          fflush(pFGSTraceFile);
        }       
#endif
        end = decodeBlockBP(vdata_ac[i/4],level,max_level);
        if(end == -1) break;
        vdata_ac[i/4]->decoded_level = max_level - level;

        for(j=0;j < 4; j++)
          rleDecoding(vdata_ac[i/4],vBuffer_ac[j],j);
      }
      
      //scanning in zigzag order the bitplane
      zigzag_scan(4*a,4*b,0,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a,4*b+2,4,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a+2,4*b,8,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a+2,4*b+2,12,yBuffer_ac,pFGSData->YBitplane_ac);

      zigzag_scan(2*a,2*b,0,uBuffer_ac,pFGSData->UBitplane_ac);
      zigzag_scan(2*a,2*b,0,vBuffer_ac,pFGSData->VBitplane_ac);

      for(c=0; c<TSIZE; c++)
      {
        pFGSData->UBitplane_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] = uBuffer_dc[c];
        pFGSData->VBitplane_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] = vBuffer_dc[c];
      }
      i+=4;
    }
    if(end == -1) break;
  }

  free_mem2D(yBuffer_ac);
  free_mem2D(uBuffer_ac);
  free_mem2D(vBuffer_ac);
  free(yBuffer_dc);
  free(uBuffer_dc);
  free(vBuffer_dc);  
  return end;
}

void zigzag_scan(int hIndex,int wIndex,int i,byte **buffer,byte **bitplane)
{
  int a,b;

  //scanning in zigzag order the bitplane to store in a lum & chrominance buffer
  for(a=0; a<TSIZE; a++)
  {
    for(b=0; b<TSIZE; b++)
    {
      bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b] = buffer[i][fgs_zigzag[a*TSIZE+b]];
      bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b] = buffer[i+1][fgs_zigzag[a*TSIZE+b]];
      bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b] = buffer[i+2][fgs_zigzag[a*TSIZE+b]];
      bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b] = buffer[i+3][fgs_zigzag[a*TSIZE+b]];
    }
  }
}

//scanning in zigzag order the bitplane to store the signal
void scanSignal(int hIndex,int wIndex,int sign[][TSIZE2],byte **bitplane)
{
  int a,b;

  for(a=0; a<TSIZE; a++)
  {
    for(b=0; b<TSIZE; b++)
    {
      bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b] = (byte)(sign[0][fgs_zigzag[a*TSIZE+b]]);
      bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b] = (byte)(sign[1][fgs_zigzag[a*TSIZE+b]]);
      bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b] = (byte)(sign[2][fgs_zigzag[a*TSIZE+b]]);
      bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b] = (byte)(sign[3][fgs_zigzag[a*TSIZE+b]]);
    }
  }
}

void recoverSignal(int hIndex,int wIndex,int value,int **bitplane)
{
  int a,b;

  for(a=0; a<TSIZE; a++)
  {
    for(b=0; b<TSIZE; b++)
    {
      if(bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b]) bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b] += value;
      if(bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b]) bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b] += value;
      if(bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b]) bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b] += value;
      if(bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b]) bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b] += value;
    }
  }
}

void finishEncoding(struct img_par *pimg)
{
  int a,b,c,d,i,j,n;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  
  enh_block **ldata_dc = pFGSData->block_data_lum_dc;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

  int decodeLevel,recover_value[6];

  i=0;j=0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      decodeLevel = ldata_ac[i]->decoded_level;
      recover_value[0] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = ldata_ac[i+1]->decoded_level;
      recover_value[1] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = ldata_ac[i+2]->decoded_level;
      recover_value[2] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = ldata_ac[i+3]->decoded_level;
      recover_value[3] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = udata_ac[j]->decoded_level;
      recover_value[4] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = vdata_ac[j]->decoded_level;
      recover_value[5] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;

      recoverSignal(4*a,4*b,recover_value[0],pFGSData->Ydiff_ac);
      recoverSignal(4*a,4*b+2,recover_value[1],pFGSData->Ydiff_ac);
      recoverSignal(4*a+2,4*b,recover_value[2],pFGSData->Ydiff_ac);
      recoverSignal(4*a+2,4*b+2,recover_value[3],pFGSData->Ydiff_ac);

      recoverSignal(2*a,2*b,recover_value[4],pFGSData->Udiff_ac);
      recoverSignal(2*a,2*b,recover_value[5],pFGSData->Vdiff_ac);

      scanSignal(4*a,4*b,ldata_ac[i]->sign,pFGSData->YSignal_ac);
      scanSignal(4*a,4*b+2,ldata_ac[i+1]->sign,pFGSData->YSignal_ac);
      scanSignal(4*a+2,4*b,ldata_ac[i+2]->sign,pFGSData->YSignal_ac);
      scanSignal(4*a+2,4*b+2,ldata_ac[i+3]->sign,pFGSData->YSignal_ac);

      scanSignal(2*a,2*b,udata_ac[j]->sign,pFGSData->USignal_ac);
      scanSignal(2*a,2*b,vdata_ac[j]->sign,pFGSData->VSignal_ac);
      i+=4;
      j++;
    }
  }

  //get the sign of each DC chrominance coefficient
  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      decodeLevel = ldata_dc[i]->decoded_level;
      recover_value[0] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = udata_dc[i]->decoded_level;
      recover_value[1] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;
      decodeLevel = vdata_dc[i]->decoded_level;
      recover_value[2] = (decodeLevel>2) ? (1<<(decodeLevel-2)) : 0;

      for(c=0; c<TSIZE; c++)
      {
        if(pFGSData->Udiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]) 
          pFGSData->Udiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] += recover_value[1];

        if(pFGSData->Vdiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]) 
          pFGSData->Vdiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] += recover_value[2];

        pFGSData->USignal_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] = udata_dc[i]->sign[0][c];
        pFGSData->VSignal_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c] = vdata_dc[i]->sign[0][c];
      }
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          if(pFGSData->Ydiff_dc[a*TSIZE+c][b*TSIZE+d]) 
            pFGSData->Ydiff_dc[a*TSIZE+c][b*TSIZE+d] += recover_value[0];

          pFGSData->YSignal_dc[a*TSIZE+c][b*TSIZE+d] = ldata_dc[i]->sign[0][fgs_zigzag[c*TSIZE+d]];
        }
      }
      i++;
    }
  }

  //apply the signal to diff
  for(a=0; a<pimg->height ; a++)
  {
    for(b=0; b<pimg->width ; b++)
    {
      if(pFGSData->YSignal_ac[a][b] == 1)
        pFGSData->Ydiff_ac[a][b] = pFGSData->Ydiff_ac[a][b]*(-1);   
    }
  }

  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      if(pFGSData->USignal_ac[a][b] == 1)
        pFGSData->Udiff_ac[a][b] = pFGSData->Udiff_ac[a][b]*(-1);

      if(pFGSData->VSignal_ac[a][b] == 1)
        pFGSData->Vdiff_ac[a][b] = pFGSData->Vdiff_ac[a][b]*(-1);
    }
  }
  
  for(a=0; a<(pimg->height>>2) ; a++)
  {
    for(b=0; b<(pimg->width>>2) ; b++)
    {
      if(pFGSData->YSignal_dc[a][b] == 1)
        pFGSData->Ydiff_dc[a][b] = pFGSData->Ydiff_dc[a][b]*(-1);
    }
  }

  n = (pimg->height_cr>>2)*(pimg->width_cr>>2); 
  
  for(a=0; a<n ; a++)
  {
    if(pFGSData->USignal_dc[a] == 1)
      pFGSData->Udiff_dc[a] = pFGSData->Udiff_dc[a]*(-1);

    if(pFGSData->VSignal_dc[a] == 1)
      pFGSData->Vdiff_dc[a] = pFGSData->Vdiff_dc[a]*(-1);
  }
}

int decodeBlockBP_lumDC(enh_block* block_data, int level)
{
  int n,end,os,sign,data,len,index,ind_level;

  end = 0;
  n = 0;
  os = 0;

  if(level > 2)
  {
    ind_level = 2;
  }
  else
  {
    ind_level = level;
  }

  while(end == 0)
  {
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 5 || data > 32)
    {
      printf("Error in decodeBlockBP_lumDC\n");
      exit(-1);
    }
    index = (int)table_dc_lum_uvlc[ind_level][data];
    
    if(index < 16)  //half of the table => eop = 0
    {
      block_data->run[0][n] = index;
      block_data->end[0][n] = 0;
    }
    else
    {
      block_data->run[0][n] = index-16;
      block_data->end[0][n] = 1;
      end = 1;
    }

#ifdef FGS_TRACE
    if(gtr == 146)
    {
      fprintf(pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][n],block_data->end[0][n]);
      fflush(pFGSTraceFile); 
    }    
#endif

    os += block_data->run[0][n]+1;
    
    if(block_data->run[0][0]==15 && block_data->end[0][0]==0) break;
    
    if (block_data->residual_sign[0][os-1] == 0)
    {
      if(readIntFGS(&sign,1,pFGSBitstream) == -1) return -1;
      block_data->sign[0][os-1] = sign;            //store the sign
      block_data->residual_sign[0][os-1] = 1;      //mark that is already read...
#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSTraceFile,"Signal:%d\n",block_data->sign[0][os-1]);
        fflush(pFGSTraceFile); 
      }
#endif
    }
    n++;
  }
  return 1;
}

int decodeBlockBP_chrDC(enh_block* block_data, int level, int max_level)
{
  int end,i,n,os,sign,index,temp,data,len;
  byte cbp;

  //clean all the data
  
  memset(block_data->end[0],0,TSIZE2*sizeof(int));
  for(i=0;i<TSIZE2;i++) block_data->run[0][i]= 3;
  end = 0;
  n = 0;
  os = 0;
  
  //reads the cbp
  if(block_data->data_level == -1)
  {
    if(level > 1)
    {
      if(readByteFGS(&cbp,1,pFGSBitstream) == -1) return -1;
      if(cbp == 1) block_data->cbp = 0;
      else block_data->cbp = 1;
    }
    if(block_data->cbp == 1)    //means that is the first time this block 'appears'
    {
      block_data->data_level = max_level - level;
      level = 0;
    }
    else return 1;                //cbp = 1 => most sgnificant block has not appeared
  }
  else
  {
    temp = block_data->data_level - (max_level-level);
    if(temp > 2)
    {
      level = 2;
    }
    else
    {
      level = temp;
    }
  }

  while(end == 0)
  {
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 3 || data > 8)
    {
      printf("Error in decodeBlockBP_chrDC\n");
      exit(-1);
    }
    index = (int)table_dc_chr_uvlc[level][data];

    if(index < 4)  //half of the table => eop = 0
    {
      block_data->run[0][n] = index;
      block_data->end[0][n] = 0;
    }
    else
    {
      block_data->run[0][n] = index-4;
      block_data->end[0][n] = 1;
      end = 1;
    }

#ifdef FGS_TRACE
    if(gtr == 146)
    {
      fprintf(pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][n],block_data->end[0][n]);
      fflush(pFGSTraceFile); 
    }
#endif

    os += block_data->run[0][n]+1;
    
    if(block_data->run[0][0]==3 && block_data->end[0][0]==0) break;
    
    if (block_data->residual_sign[0][os-1] == 0)
    {
      if(readIntFGS(&sign,1,pFGSBitstream) == -1) return -1;
      block_data->sign[0][os-1] = sign;            //store the sign
      block_data->residual_sign[0][os-1] = 1;      //mark that is already read...
#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSTraceFile,"Signal:%d\n",block_data->sign[0][os-1]);
        fflush(pFGSTraceFile); 
      }
#endif
    }
    n++;
  }
  return 1;
}

//verificar se o 'os' est fixe
int decodeBlockBP(enh_block* block_data, int level, int max_level)
{
  int end,i,j,temp,n,os,index,sign,data,len,ind_table;
  byte cbp;

  //clean all the data
  for(j=0;j<4;j++)
  {
    memset(block_data->end[j],0,TSIZE2*sizeof(int));
    for(i=0;i<TSIZE;i++) block_data->run[j][i]= 15;
  }

  //reads the cbp
  if(block_data->data_level == -1)
  {
    if(level > 1)
    {
      if(readByteFGS(&cbp,1,pFGSBitstream) == -1) return -1;
      if(cbp == 1) block_data->cbp = 0;
      else block_data->cbp = 1;
    }
    if(block_data->cbp == 1)    //means that is the first time this block 'appears'
    {
      block_data->data_level = max_level - level;
      level = 0;
    }
    else return 1;                //cbp = 1 => most sgnificant block has not appeared
  }
  else
  {
    temp = block_data->data_level - (max_level-level);
    if(temp > 2)
    {
      level = 2;
    }
    else
    {
      level = temp;
    }
  }

  for(index = 0;index < 4;index++)
  {
    n = 0;os = 0;end=0;
    while(end == 0)
    {
      len = decodeVLCCode(&data);
      if(len == -1) return -1;
      if(len > 5 || data > 32)
      {
        printf("Error in decodeBlockBP\n");
        exit(-1);
      }
      ind_table = (int)table_ac_uvlc[level][data];

      if(ind_table < 16)  //half of the table => eop = 0
      {
        block_data->run[index][n] = ind_table;
        block_data->end[index][n] = 0;
      }
      else
      {
        block_data->run[index][n] = ind_table-16;
        block_data->end[index][n] = 1;
        end = 1;
      }

#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[index][n],block_data->end[index][n]);
        fflush(pFGSTraceFile); 
      }
#endif

      os += block_data->run[index][n]+1;

      if(block_data->run[index][0]==15 && block_data->end[index][0]==0) break;
      
      if (block_data->residual_sign[index][os-1] == 0)
      {
        if(readIntFGS(&sign,1,pFGSBitstream) == -1) return -1;
        block_data->sign[index][os-1] = sign;            //store the sign
        block_data->residual_sign[index][os-1] = 1;      //mark that is already read...
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSTraceFile,"Signal:%d\n",block_data->sign[index][os-1]);
          fflush(pFGSTraceFile); 
        }
#endif
      }
      n++;
    }
  }
  return 1;
}

int decodeVLCCode(int *value)
{
  byte bit;
  int n,ret;
  n = 0;

  do{
    ret = readByteFGS(&bit,1,pFGSBitstream);
    if(ret == -1) return -1;
    n++;
  }while(bit == 0);

  if(n > 6)
  {
    printf("Error in decodeVLCCode\n");
    exit(-1);
  }
  if(n>1)
  {
    ret = readIntFGS(value,n-1,pFGSBitstream);
    if(ret == -1) return -1;
    *value = *value | (1<<(n-1));     //put back the middle 1
    return n-1;
  }
  else
  {
    *value = 1;
    return 1;
  }
}

int getMBIndex(int level)
{
  int index,yy_ac,yy_dc,dc_chroma,ac_chroma;
  yy_ac = pFGSData->maximum_level_y_ac;
  yy_dc = pFGSData->maximum_level_y_dc;
  dc_chroma = pFGSData->maximum_level_chr_dc; //max(pFGSData->maximum_level_u_dc,pFGSData->maximum_level_v_dc);
  ac_chroma = pFGSData->maximum_level_chr_ac; //max(pFGSData->maximum_level_u_ac,pFGSData->maximum_level_v_ac);
  
  if (yy_dc >= level && yy_ac < level && dc_chroma < level && ac_chroma < level)  //only dc coeficients are coded in this vop-bp
  {
    index = -1;
  }
  else if (yy_ac >= level && dc_chroma < level && ac_chroma < level)  //nc = 0
  {
    index = 0;
  }
  else if (yy_ac >= level && dc_chroma >= level && ac_chroma < level)  //nc = 1
  {
    index = 1;
  }
  else if (yy_ac >= level && ac_chroma >= level) //nc = 2
  {
    index = 2;
  }
  else if (yy_ac < level && dc_chroma >= level && ac_chroma < level) //nc = 1
  {
    index = 3;
  }
  else if (yy_ac < level && ac_chroma >= level) //nc = 2
  {
    index = 4;
  }
  else
  {
    printf("Error in the getMBIndex\n");
    exit(1);
  }
  return(index);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the run lenght decoding in a block
 * \param block_data
 *    pointer to the input enh_block structure 
 * \param out
 *    pointer to input buffer
 * \param level
 *    bitplane level
 * \return
 *    none
 ***********************************************************************
 */

void rleDecoding(enh_block* block_data,byte *out,int t_block)
{
  int i, j, n, pos;
  
  memset(out,0,TSIZE2);
  
  if (block_data->run[t_block][0] == 15 && block_data->end[t_block][0] == 0 ) 
  { 
    n = 0;
  } 
  else 
  {
    n = 0;
    while ( block_data->end[t_block][n] == 0 && n < 16 ) n += 1;
    if ( n >= 16) 
    {
      printf("Wrong RLE code: n = %d\n",n);
      exit(1);
    }
    n += 1;
  }
  //if n == 0 means that all is all zeros =>run = 63 & end = 0
  if ( n!= 0) 
  {
    pos = 0;
    for (i = 0; i < n; i++)
    {
      for (j=0; j < block_data->run[t_block][i]; j++ ) out[pos+j] = 0;
      out[pos+j] = 1;
      pos += block_data->run[t_block][i]+1;
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    performs the run lenght decoding in a block
 * \param block_data
 *    pointer to the input enh_block structure 
 * \param out
 *    pointer to input buffer
 * \param level
 *    bitplane level
 * \return
 *    none
 ***********************************************************************
 */

void rleDecoding2x2(enh_block* block_data,byte *out,int t_block)
{
  int i, j, n, pos;
  
  memset(out,0,4);
  
  if (block_data->run[t_block][0] == 3 && block_data->end[t_block][0] == 0 ) 
  { 
    n = 0;
  } 
  else 
  {
    n = 0;
    while ( block_data->end[t_block][n] == 0 && n < 4 ) n += 1;
    if ( n >= 4) 
    {
      printf("Wrong RLE code: n = %d\n",n);
      exit(1);
    }
    n += 1;
  }
  //if n == 0 means that all is all zeros =>run = 63 & end = 0
  if ( n!= 0) 
  {
    pos = 0;
    for (i = 0; i < n; i++)
    {
      for (j=0; j < block_data->run[t_block][i]; j++ ) out[pos+j] = 0;
      out[pos+j] = 1;
      pos += block_data->run[t_block][i]+1;
    }
  }
}

int readCbpindexMSB(int comb_index,int block_index,byte *except_code)
{
  int b,data,len;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

  switch(comb_index) 
  {
  case 0:
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 4 || data > 16)
    {
      printf("Error in readCbpindexMSB\n");
      exit(-1);
    }
    b = (int)table_cbp_yyyy[0][data];
    ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & b;
  break;
  case 1:
  case 2:
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 5 || data > 48)
    {
      printf("Error in readCbpindexMSB\n");
      exit(-1);
    }
    b = (int)table_cbp_yyyyuv[0][data];
    ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & b;
    if((b>>4) == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if((b>>4) == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if((b>>4) == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
  break;
  case 3:
  case 4:
    peekIntFGS(&data,1,pFGSBitstream);
    if(data == 1)
    {
      readIntFGS(&data,1,pFGSBitstream);
      b = 0;
    }
    else
    {
      readIntFGS(&data,1,pFGSBitstream);
      readIntFGS(&b,2,pFGSBitstream);
    }
    if(b == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if(b == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if(b == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
  break;
  default:
    printf("Error in the readCbpIndexMSB!\n");
    exit(1);
  }
  if(b == 0) except_code[block_index/4] = 1;

#ifdef FGS_TRACE
  if(gtr == 146)
  {
    fprintf(pFGSTraceFile,"CombIndex: %d  CbpIndex: %d\n",comb_index,b);
    fflush(pFGSTraceFile); 
  }
#endif

  return 1;
}

int readCbpindex(int comb_index,int block_index,byte *except_code)
{
  int b,data,len;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

  if(except_code[block_index/4] == 1)
  {
    switch(comb_index) 
    {
    case 0:
      len = decodeVLCCode(&data);
      if(len == -1) return -1;
      if(len > 4 || data > 16)
      {
        printf("Error in readCbpindex\n");
        exit(-1);
      }
      b = (int)table_cbp_yyyy[1][data];
      ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
      ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
      ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
      ldata_ac[block_index+3]->cbp = 0x1 & b;
      break;
    case 1:
    case 2:
      len = decodeVLCCode(&data);
      if(len == -1) return -1;
      if(len > 5 || data > 48)
      {
        printf("Error in readCbpindex\n");
        exit(-1);
      }
      b = (int)table_cbp_yyyyuv[1][data];
      ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
      ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
      ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
      ldata_ac[block_index+3]->cbp = 0x1 & b;
      if((b>>4) == 0)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x00;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x00;
      }
      if((b>>4) == 1)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      if((b>>4) == 2)
      {
        udata_ac[block_index/4]->cbp = 0x01;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x01;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      break;
    case 3:
    case 4:
      readIntFGS(&b,2,pFGSBitstream);
      if(b == 0)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x00;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x00;
      }
      if(b == 1)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      if(b == 2)
      {
        udata_ac[block_index/4]->cbp = 0x01;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x01;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      break;
    default:
      printf("Error in the readCbpIndexM!\n");
      exit(1);
    }
#ifdef FGS_TRACE
    if(gtr == 146)
    {
      fprintf(pFGSTraceFile,"CombIndex: %d  CbpIndex: %d  ExceptCode: %d\n",comb_index,b,except_code[block_index/4]);
      fflush(pFGSTraceFile); 
    }
#endif
    return 1;
  }
  switch(comb_index) 
  {
  case 0:
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 4 || data > 16)
    {
      printf("Error in readCbpindex\n");
      exit(-1);
    }
    b = (int)table_cbp_yyyy[2][data];
    ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & b;
    break;
  case 1:
  case 2:
    len = decodeVLCCode(&data);
    if(len == -1) return -1;
    if(len > 5 || data > 48)
    {
      printf("Error in readCbpindex\n");
      exit(-1);
    }
    b = (int)table_cbp_yyyyuv[2][data];
    ldata_ac[block_index]->cbp = 0x1 & (b >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (b >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (b >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & b;
    if((b>>4) == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if((b>>4) == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if((b>>4) == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    break;
  case 3:
  case 4:
    readIntFGS(&b,2,pFGSBitstream);
    if(b == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if(b == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if(b == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    break;
  default:
    printf("Error in the readCbpIndex!\n");
    exit(1);
  }
#ifdef FGS_TRACE
  if(gtr == 146)
  {
    fprintf(pFGSTraceFile,"CombIndex: %d  CbpIndex: %d  ExceptCode: %d\n",comb_index,b,except_code[block_index/4]);
    fflush(pFGSTraceFile); 
  }
#endif
  return 1;
}

void calcBitplane(struct img_par *pimg,int level)
{
  int a,b,n;

  printf("Level %d decoded\n",level);
 
  if(pFGSData->skip_y_ac == FALSE)
  { 
    for(a=0; a<pimg->height ; a++)
    {
      for(b=0; b<pimg->width ; b++)
      {
        pFGSData->Ydiff_ac[a][b] = (pFGSData->Ydiff_ac[a][b] << 1) + pFGSData->YBitplane_ac[a][b];
        pFGSData->YBitplane_ac[a][b] = 0;
      }
    }
  }

  if(pFGSData->skip_chr_ac == FALSE)
  { 
    for(a=0; a<pimg->height_cr ; a++)
    {
      for(b=0; b<pimg->width_cr ; b++)
      {
        pFGSData->Udiff_ac[a][b] = (pFGSData->Udiff_ac[a][b] << 1) + (pFGSData->UBitplane_ac[a][b]);
        pFGSData->Vdiff_ac[a][b] = (pFGSData->Vdiff_ac[a][b] << 1) + (pFGSData->VBitplane_ac[a][b]);
        pFGSData->UBitplane_ac[a][b] = 0;
        pFGSData->VBitplane_ac[a][b] = 0;
      }
    }
  }

  //DC components
  if(pFGSData->skip_y_dc == FALSE)
  { 
    for(a=0; a<(pimg->height>>2) ; a++)
    {
      for(b=0; b<(pimg->width>>2) ; b++)
      {
        pFGSData->Ydiff_dc[a][b] = (pFGSData->Ydiff_dc[a][b] << 1) + pFGSData->YBitplane_dc[a][b];
        pFGSData->YBitplane_dc[a][b] = 0;
      }
    }
  }

  n = (pimg->height_cr>>2)*(pimg->width_cr>>2); 
  if(pFGSData->skip_chr_dc == FALSE)
  { 
    for(a=0; a<n ; a++)
    {
      pFGSData->Udiff_dc[a] = (pFGSData->Udiff_dc[a] << 1) + (pFGSData->UBitplane_dc[a]);
      pFGSData->Vdiff_dc[a] = (pFGSData->Vdiff_dc[a] << 1) + (pFGSData->VBitplane_dc[a]);
      pFGSData->UBitplane_dc[a] = 0;
      pFGSData->VBitplane_dc[a] = 0;
    }
  }
  //put the bitplanes with zeros ... since they have the values of the last level ...
  /*if(pFGSData->skip_chr_ac == FALSE && pFGSData->skip_chr_dc == TRUE)
  {
    memset(pFGSData->UBitplane_dc,0,n*sizeof(byte));
    memset(pFGSData->VBitplane_dc,0,n*sizeof(byte));
  }*/
}

void endCalcBitplane(struct img_par *pimg,int nBp)
{
  int a,b,n;
  if(nBp < 1) return;
 
  for(a=0; a<pimg->height ; a++)
  {
    for(b=0; b<pimg->width ; b++)
    {
      pFGSData->Ydiff_ac[a][b] = (pFGSData->Ydiff_ac[a][b] << nBp);
    }
  }
  
  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      pFGSData->Udiff_ac[a][b] = (pFGSData->Udiff_ac[a][b] << nBp);
      pFGSData->Vdiff_ac[a][b] = (pFGSData->Vdiff_ac[a][b] << nBp);
    }
  }
  for(a=0; a<(pimg->height>>2) ; a++)
  {
    for(b=0; b<(pimg->width>>2) ; b++)
    {
      pFGSData->Ydiff_dc[a][b] = (pFGSData->Ydiff_dc[a][b] << nBp);
    }
  }
  n = (pimg->height_cr>>2)*(pimg->width_cr>>2); 
  for(a=0; a<n ; a++)
  {
    pFGSData->Udiff_dc[a] = (pFGSData->Udiff_dc[a] << nBp);
    pFGSData->Vdiff_dc[a] = (pFGSData->Vdiff_dc[a] << nBp);
  }
}

void clearData(struct img_par *pimg)
{
  int a,b,n;
  for(a=0; a<pimg->height ; a++)
  {
    for(b=0; b<pimg->width ; b++)
    {
      pFGSData->Ydiff_ac[a][b] = 0;
    }
  }

  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      pFGSData->Udiff_ac[a][b] = 0;
      pFGSData->Vdiff_ac[a][b] = 0;
    }
  }

  for(a=0; a<(pimg->height>>2) ; a++)
  {
    for(b=0; b<(pimg->width>>2) ; b++)
    {
      pFGSData->Ydiff_dc[a][b] = 0;
    }
  }
  

  n = (pimg->height_cr>>2)*(pimg->width_cr>>2); 
  for(a=0; a<n ; a++)
  {
    pFGSData->Udiff_dc[a] = 0;
    pFGSData->Vdiff_dc[a] = 0;
  }
}

void addResidueToImg(struct img_par *pimg)
{
  int i,j,a,b;
  
  for(i=0;i<pimg->height;i++)
    for(j=0;j<pimg->width;j++)
    {
      a = pFGSData->Ydiff_ac[i][j] + imgY[i][j];
      if(a > 255) a = 255;
      if(a < 0) a = 0;
      fgs_imgY[i][j] = a;
    }

  for(i=0;i<pimg->height_cr;i++)
    for(j=0;j<pimg->width_cr;j++)
    {
      a = pFGSData->Udiff_ac[i][j] + imgUV[0][i][j];
      b = pFGSData->Vdiff_ac[i][j] + imgUV[1][i][j];
      if(a > 255) a = 255;
      if(a < 0) a = 0;
      if(b > 255) b = 255;
      if(b < 0) b = 0;
      fgs_imgUV[0][i][j] = a;
      fgs_imgUV[1][i][j] = b;
    }
}

void writeFGSFrame(struct img_par *img)
{
  int i,j;

  for(i=0;i<img->height;i++)
    for(j=0;j<img->width;j++)
    {
      fputc(fgs_imgY[i][j],pFGSout);
    }
  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
    {
      fputc(fgs_imgUV[0][i][j],pFGSout);
    }
  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
    {
      fputc(fgs_imgUV[1][i][j],pFGSout);
    }
}

void writePrevFGSFrame(struct img_par *img)
{
  int i,j;
  
  for(i=0;i<img->height;i++)
    for(j=0;j<img->width;j++)
      fputc(fgs_imgY_prev[i][j],pFGSout);

  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
      fputc(fgs_imgUV_prev[0][i][j],pFGSout);

  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
      fputc(fgs_imgUV_prev[1][i][j],pFGSout);

  fprintf(stdout,"\nFGS decoded frame written\n");
}

void copyFGSframe(struct img_par *img)
{
  int i,j;

  for(i=0;i<img->height;i++)
    for(j=0;j<img->width;j++)
    {
      fgs_imgY_prev[i][j] = fgs_imgY[i][j];
    }
  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
    {
      fgs_imgUV_prev[0][i][j] = fgs_imgUV[0][i][j];
    }
  for(i=0;i<img->height_cr;i++)
    for(j=0;j<img->width_cr;j++)
    {
      fgs_imgUV_prev[1][i][j] = fgs_imgUV[1][i][j];
    }
}

void cleanFGS()
{
  int a,nblocks;
  fclose(pFGSout);
#ifdef FGS_TRACE
  fclose(pFGSTraceFile);
#endif
  closeBitstream();
  //free residual images
  free_mem2Dint(pFGSData->Ydiff_ac);
  free_mem2Dint(pFGSData->Udiff_ac);
  free_mem2Dint(pFGSData->Vdiff_ac);
  free_mem2Dint(pFGSData->Ydiff_dc);
  free(pFGSData->Udiff_dc);
  free(pFGSData->Vdiff_dc);
  
  //free bitplanes
  free_mem2D(pFGSData->YBitplane_ac);
  free_mem2D(pFGSData->UBitplane_ac);
  free_mem2D(pFGSData->VBitplane_ac);
  free_mem2D(pFGSData->YBitplane_dc);
  free(pFGSData->UBitplane_dc);
  free(pFGSData->VBitplane_dc);
  //free signal array
  free_mem2D(pFGSData->YSignal_ac);
  free_mem2D(pFGSData->USignal_ac);
  free_mem2D(pFGSData->VSignal_ac);
  free_mem2D(pFGSData->YSignal_dc);
  free(pFGSData->USignal_dc);
  free(pFGSData->VSignal_dc);
  
  //free structure with encoding data for luminance
  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;
  for(a=0;a < (4*nblocks);a++)
  {
    free(pFGSData->block_data_lum_ac[a]);
  }
  free(pFGSData->block_data_lum_ac);
  for(a=0;a < nblocks;a++)
  {
    free(pFGSData->block_data_lum_dc[a]);
  }
  free(pFGSData->block_data_lum_dc);
  
  //free structure with encoding data for chrominance
  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;
  for(a=0;a < nblocks;a++)
  {
    free(pFGSData->block_data_u_ac[a]);
    free(pFGSData->block_data_v_ac[a]);
    free(pFGSData->block_data_u_dc[a]);
    free(pFGSData->block_data_v_dc[a]);
  }
  free(pFGSData->block_data_u_ac);
  free(pFGSData->block_data_v_ac);
  free(pFGSData->block_data_u_dc);
  free(pFGSData->block_data_v_dc);

}

/*!
 ***********************************************************************
 * \brief
 *    Inverse 4x4 transformation
 ***********************************************************************
 */
void H26L_itransLum(int **coefAC,int **coefDC)
{
  int i,ii,jj,j,i1,j1;
  int m5[4];
  int m6[4];
  int m0[4][4][4][4];

  //put in index,block_num,index,block_num 4 dimensional array
  for (j=0;j<16;j++)
  {
    for (i=0;i<16;i++)
    {
      m0[i%4][i/4][j%4][j/4]=coefAC[i][j];
    }
  }

  //hadamard Itransform of DC coeficients
  
  for (j=0;j<4;j++)
  {
    for (i=0;i<4;i++)
      m5[i]=coefDC[i][j];

    m6[0]=m5[0]+m5[2];
    m6[1]=m5[0]-m5[2];
    m6[2]=m5[1]-m5[3];
    m6[3]=m5[1]+m5[3];

    for (i=0;i<2;i++)
    {
      i1=3-i;
      coefDC[i][j]=m6[i]+m6[i1];
      coefDC[i1][j]=m6[i]-m6[i1];
    }
  }

  // vertical
  for (i=0;i<4;i++)
  {
    for (j=0;j<4;j++)
      m5[j]=coefDC[i][j];

    m6[0]=m5[0]+m5[2];
    m6[1]=m5[0]-m5[2];
    m6[2]=m5[1]-m5[3];
    m6[3]=m5[1]+m5[3];

    for (j=0;j<2;j++)
    {
      j1=3-j;
      coefDC[i][j] = m6[j]+m6[j1];
      coefDC[i][j1] = m6[j]-m6[j1];
    }
  }

  //scaling of the DC coeficients (Q=0)
  for (j=0;j<4;j++)
  {
    for (i=0;i<4;i++)
    {
      coefDC[i][j] = (coefDC[i][j]*fgs_dequant_coef[0][0])/4;
    }
  }

  //scaling of the AC coeficients (Q=0)
  for (j=0;j<16;j++)
  {
    for (i=0;i<16;i++)
    {
        m0[i%4][i/4][j%4][j/4]=coefAC[i][j]*fgs_dequant_coef[i%4][j%4];
    }
  }

  //put back the DC coeficients
  for (j=0;j<4;j++)
    for (i=0;i<4;i++)
      m0[0][i][0][j]= coefDC[i][j];

  //IDCT

  for (jj=0;jj<4;jj++)
  {
    for (ii=0;ii<4;ii++)
    {
      // horizontal
      for (j=0;j<4;j++)
      {
        for (i=0;i<4;i++)
        {
          m5[i]=m0[i][ii][j][jj];
        }
        m6[0]=(m5[0]+m5[2]);
        m6[1]=(m5[0]-m5[2]);
        m6[2]=(m5[1]>>1)-m5[3];
        m6[3]=m5[1]+(m5[3]>>1);
        
        for (i=0;i<2;i++)
        {
          i1=3-i;
          m0[i][ii][j][jj]=m6[i]+m6[i1];
          m0[i1][ii][j][jj]=m6[i]-m6[i1];
        }
      }
      //vertical   
      for (i=0;i<BLOCK_SIZE;i++)
      {
        for (j=0;j<BLOCK_SIZE;j++)
        {
          m5[j]=m0[i][ii][j][jj];
        }
        m6[0]=(m5[0]+m5[2]);
        m6[1]=(m5[0]-m5[2]);
        m6[2]=(m5[1]>>1)-m5[3];
        m6[3]=m5[1]+(m5[3]>>1);
        
        for (j=0;j<2;j++)
        {
          j1=3-j;
          m0[i][ii][j][jj]=m6[j]+m6[j1];
          m0[i][ii][j1][jj]=m6[j]-m6[j1];
        }
      }
      //end of 4x4 block IDCT
    }
  }

  //put back the AC coeficients
  for (j=0;j<16;j++)
  {
    for (i=0;i<16;i++)
    {
      coefAC[i][j]=(m0[i%4][i/4][j%4][j/4]+(1<<5))>>6;
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    Inverse 4x4 transformation
 ***********************************************************************
 */
void H26L_itransChr(int **coefAC,int *coefDC)
{
  int i,j,ii,jj,i1,j1;
  int m5[4];
  int m6[4];
  int m0[4][4][4][4];

  //put in index,block_num,index,block_num 4 dimensional array

  for (j=0;j<8;j++)
  {
    for (i=0;i<8;i++)
    {
      m0[i%4][i/4][j%4][j/4]=coefAC[i][j];
    }
  }

  //scaling of the DC coeficients (Q=0)
  for (i=0;i<4;i++)
  {
    coefDC[i] = (coefDC[i]*fgs_dequant_coef[0][0]);
  }
  //scaling of the AC coeficients (Q=0)
  for (j=0;j<8;j++)
  {
    for (i=0;i<8;i++)
    {
      m0[i%4][i/4][j%4][j/4]=coefAC[i][j]*fgs_dequant_coef[i%4][j%4];
    }
  }

  m0[0][0][0][0]=(coefDC[0]+coefDC[1]+coefDC[2]+coefDC[3])/2;
  m0[0][1][0][0]=(coefDC[0]-coefDC[1]+coefDC[2]-coefDC[3])/2;
  m0[0][0][0][1]=(coefDC[0]+coefDC[1]-coefDC[2]-coefDC[3])/2;
  m0[0][1][0][1]=(coefDC[0]-coefDC[1]-coefDC[2]+coefDC[3])/2;

  for (jj=0;jj<2;jj++)
  {
    for (ii=0;ii<2;ii++)
    {
      // horizontal
      for (j=0;j<4;j++)
      {
        for (i=0;i<4;i++)
        {
          m5[i]=m0[i][ii][j][jj];
        }
        m6[0]=(m5[0]+m5[2]);
        m6[1]=(m5[0]-m5[2]);
        m6[2]=(m5[1]>>1)-m5[3];
        m6[3]=m5[1]+(m5[3]>>1);
        
        for (i=0;i<2;i++)
        {
          i1=3-i;
          m0[i][ii][j][jj]=m6[i]+m6[i1];
          m0[i1][ii][j][jj]=m6[i]-m6[i1];
        }
      }
      //vertical   
      for (i=0;i<4;i++)
      {
        for (j=0;j<4;j++)
        {
          m5[j]=m0[i][ii][j][jj];
        }
        m6[0]=(m5[0]+m5[2]);
        m6[1]=(m5[0]-m5[2]);
        m6[2]=(m5[1]>>1)-m5[3];
        m6[3]=m5[1]+(m5[3]>>1);
        
        for (j=0;j<2;j++)
        {
          j1=3-j;
          m0[i][ii][j][jj]=(m6[j]+m6[j1]);
          m0[i][ii][j1][jj]=(m6[j]-m6[j1]);
        }
      }
      //end of 4x4 block IDCT
    }
  }

  //put back the AC coeficients
  for (j=0;j<8;j++)
  {
    for (i=0;i<8;i++)
    {
      coefAC[i][j]=(m0[i%4][i/4][j%4][j/4]+(1<<5))>>6;
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    performs the inverse discrete cosine transform in an image
 * \param data
 *    pointer to matrix image data
 * \param hDCTBlocks
 *    number of blocks relative to the height of the image (vertical blocks)
 * \param wDCTBlocks
 *    number of blocks relative to the width of the image (horizontal blocks)
 * \para Notes
 *    the first cycle must be speed up with memcpy
 * \return
 *    none
 ***********************************************************************
 */

void calcitrans_lum(int ** dataAC, int ** dataDC,int hDCTBlocks,int wDCTBlocks)
{
  int a,b,c,d;
  int **ACBuffer,**DCBuffer;
   //temporary buffer for MB store
  get_mem2Dint(&ACBuffer, MBSIZE, MBSIZE);
  get_mem2Dint(&DCBuffer, BSIZE, BSIZE);

  for(a=0; a<hDCTBlocks; a++)
  {
    for(b=0; b<wDCTBlocks; b++)
    {
      for(c=0; c<MBSIZE; c++)
      {
        for(d=0; d<MBSIZE; d++)
        {
          ACBuffer[c][d] = dataAC[a*MBSIZE+c][b*MBSIZE+d];
        }
      }
      for(c=0; c<4; c++)
      {
        for(d=0; d<4; d++)
        {
          DCBuffer[c][d] = dataDC[a*4+c][b*4+d];
        }
      }
      H26L_itransLum(ACBuffer,DCBuffer);
      for(c=0; c<MBSIZE; c++)
      {
        for(d=0; d<MBSIZE; d++)
        {
          dataAC[a*MBSIZE+c][b*MBSIZE+d] = ACBuffer[c][d];
        }
      }
    }
  }
  free_mem2Dint(ACBuffer);
  free_mem2Dint(DCBuffer);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the forward discrete cosine transform in an image
 * \param data
 *    pointer to matrix image data
 * \param hDCTBlocks
 *    number of blocks relative to the height of the image (vertical blocks)
 * \param wDCTBlocks
 *    number of blocks relative to the width of the image (horizontal blocks)
 * \para Notes
 *    the first cycle must be speed up with memcpy
 * \return
 *    none
 ***********************************************************************
 */

void calcitrans_chr(int ** dataAC, int * dataDC,int hDCTBlocks,int wDCTBlocks)
{
  int a,b,c,d;
  int **ACBuffer;
  int DCBuffer[4];
   //temporary buffer for MB store
  get_mem2Dint(&ACBuffer, BSIZE, BSIZE);

  for(a=0; a<hDCTBlocks; a++)
  {
    for(b=0; b<wDCTBlocks; b++)
    {
      for(c=0; c<BSIZE; c++)
      {
        for(d=0; d<BSIZE; d++)
        {
          ACBuffer[c][d] = dataAC[a*BSIZE+c][b*BSIZE+d];
        }
      }
      
      for(c=0; c<4; c++)
      {
        DCBuffer[c] = dataDC[a*4*wDCTBlocks+b*4+c];
      }
      H26L_itransChr(ACBuffer,DCBuffer);
      for(c=0; c<BSIZE; c++)
      {
        for(d=0; d<BSIZE; d++)
        {
          dataAC[a*BSIZE+c][b*BSIZE+d] = ACBuffer[c][d];
        }
      }
    }
  }
  free_mem2Dint(ACBuffer);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the inverse discrete cosine transform in a block
 * \param data
 *    pointer to matrix with 8x8 values (the size is not checked!!)
 * \para Notes
 *    based on the integer implementation of the independent jpeg group
 * \return
 *    none
 ***********************************************************************
 */
void idctInteger (int **data)
{
  int tmp0, tmp1, tmp2, tmp3;
  int tmp10, tmp11, tmp12, tmp13;
  int z1, z2, z3, z4, z5;
  int rowctr,i;
 
  i = 0;
  for (rowctr = DCTSIZE-1; rowctr >= 0; rowctr--) 
  {
    if ((data[i][1] | data[i][2] | data[i][3] | data[i][4] |
      data[i][5] | data[i][6] | data[i][7]) == 0) 
    {
      
      int dcval = (int) (data[i][0] << PASS1_BITS);
      
      data[i][0] = dcval;
      data[i][1] = dcval;
      data[i][2] = dcval;
      data[i][3] = dcval;
      data[i][4] = dcval;
      data[i][5] = dcval;
      data[i][6] = dcval;
      data[i][7] = dcval;
      
      i++;
      continue;
    }
    
    z2 = (int) data[i][2];
    z3 = (int) data[i][6];
    
    z1 = MULTIPLY(z2 + z3, FIX_0_541196100);
    tmp2 = z1 + MULTIPLY(z3, - FIX_1_847759065);
    tmp3 = z1 + MULTIPLY(z2, FIX_0_765366865);
    
    tmp0 = ((int) data[i][0] + (int) data[i][4]) << CONST_BITS;
    tmp1 = ((int) data[i][0] - (int) data[i][4]) << CONST_BITS;
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    tmp0 = (int) data[i][7];
    tmp1 = (int) data[i][5];
    tmp2 = (int) data[i][3];
    tmp3 = (int) data[i][1];
    
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp0 = MULTIPLY(tmp0, FIX_0_298631336);
    tmp1 = MULTIPLY(tmp1, FIX_2_053119869);
    tmp2 = MULTIPLY(tmp2, FIX_3_072711026);
    tmp3 = MULTIPLY(tmp3, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    
    data[i][0] = (int) DESCALE(tmp10 + tmp3, CONST_BITS-PASS1_BITS);
    data[i][7] = (int) DESCALE(tmp10 - tmp3, CONST_BITS-PASS1_BITS);
    data[i][1] = (int) DESCALE(tmp11 + tmp2, CONST_BITS-PASS1_BITS);
    data[i][6] = (int) DESCALE(tmp11 - tmp2, CONST_BITS-PASS1_BITS);
    data[i][2] = (int) DESCALE(tmp12 + tmp1, CONST_BITS-PASS1_BITS);
    data[i][5] = (int) DESCALE(tmp12 - tmp1, CONST_BITS-PASS1_BITS);
    data[i][3] = (int) DESCALE(tmp13 + tmp0, CONST_BITS-PASS1_BITS);
    data[i][4] = (int) DESCALE(tmp13 - tmp0, CONST_BITS-PASS1_BITS);
    
    i++;
  }
  i = 0;

  for (rowctr = DCTSIZE-1; rowctr >= 0; rowctr--) {
    
    z2 = (int) data[2][i];
    z3 = (int) data[6][i];
    
    z1 = MULTIPLY(z2 + z3, FIX_0_541196100);
    tmp2 = z1 + MULTIPLY(z3, - FIX_1_847759065);
    tmp3 = z1 + MULTIPLY(z2, FIX_0_765366865);
    
    tmp0 = ((int) data[0][i] + (int) data[4][i]) << CONST_BITS;
    tmp1 = ((int) data[0][i] - (int) data[4][i]) << CONST_BITS;
    
    tmp10 = tmp0 + tmp3;
    tmp13 = tmp0 - tmp3;
    tmp11 = tmp1 + tmp2;
    tmp12 = tmp1 - tmp2;
    
    tmp0 = (int) data[7][i];
    tmp1 = (int) data[5][i];
    tmp2 = (int) data[3][i];
    tmp3 = (int) data[1][i];
    
    z1 = tmp0 + tmp3;
    z2 = tmp1 + tmp2;
    z3 = tmp0 + tmp2;
    z4 = tmp1 + tmp3;
    z5 = MULTIPLY(z3 + z4, FIX_1_175875602);
    
    tmp0 = MULTIPLY(tmp0, FIX_0_298631336);
    tmp1 = MULTIPLY(tmp1, FIX_2_053119869);
    tmp2 = MULTIPLY(tmp2, FIX_3_072711026);
    tmp3 = MULTIPLY(tmp3, FIX_1_501321110);
    z1 = MULTIPLY(z1, - FIX_0_899976223);
    z2 = MULTIPLY(z2, - FIX_2_562915447);
    z3 = MULTIPLY(z3, - FIX_1_961570560);
    z4 = MULTIPLY(z4, - FIX_0_390180644);
    
    z3 += z5;
    z4 += z5;
    
    tmp0 += z1 + z3;
    tmp1 += z2 + z4;
    tmp2 += z2 + z3;
    tmp3 += z1 + z4;
    
    data[0][i] = (int) DESCALE(tmp10 + tmp3,
      CONST_BITS+PASS1_BITS+3);
    data[7][i] = (int) DESCALE(tmp10 - tmp3,
      CONST_BITS+PASS1_BITS+3);
    data[1][i] = (int) DESCALE(tmp11 + tmp2,
      CONST_BITS+PASS1_BITS+3);
    data[6][i] = (int) DESCALE(tmp11 - tmp2,
      CONST_BITS+PASS1_BITS+3);
    data[2][i] = (int) DESCALE(tmp12 + tmp1,
      CONST_BITS+PASS1_BITS+3);
    data[5][i] = (int) DESCALE(tmp12 - tmp1,
      CONST_BITS+PASS1_BITS+3);
    data[3][i] = (int) DESCALE(tmp13 + tmp0,
      CONST_BITS+PASS1_BITS+3);
    data[4][i] = (int) DESCALE(tmp13 - tmp0,
      CONST_BITS+PASS1_BITS+3);
    i++;
  }
}

/*!
 ***********************************************************************
 * \brief
 *    initializes the bitstream
 * \param inp
 *    input parameters
 * \return
 *    none
 ***********************************************************************
 */

void initializeBitstream(struct inp_par *inp)
{
  pFGSBitstream = (decBitstream *) calloc(1, sizeof(decBitstream));
  //size of the decoding buffer
  pFGSBitstream->buffer_size = 30000;
  pFGSBitstream->streamBuffer = (byte *) calloc(pFGSBitstream->buffer_size, sizeof(byte));
  if (pFGSBitstream == NULL || pFGSBitstream->streamBuffer == NULL)
  {
    snprintf(errortext, ET_SIZE, "initializeBitstream: Memory allocation for FGS bitstream buffer failed");
    error(errortext, 100);
  }
  pFGSBitstream->byte_pos = 0;
  pFGSBitstream->bit_pos = 8;
  pFGSBitstream->p_frame = 0;

  if ((pFGSBitstream->pFGSFile=fopen(inp->FGSBitstreamFile,"rb"))==NULL)
  {
    snprintf (errortext, ET_SIZE, "initializeBitstream: Cannot open FGS bitstream file '%s'", inp->FGSBitstreamFile);
    error(errortext,500);
  }
#ifdef FGS_TRACE 
  //open trace file
  if(NULL == (pFGSTraceFile=fopen(FGS_TRACE_PATH, "w")))
  {
    snprintf (errortext, ET_SIZE, "Cannot open fgs trace file\n");
    error(errortext,500);
  }
#endif
  pFGSBitstream->read_bytes = fread(pFGSBitstream->streamBuffer, sizeof(byte), 
    pFGSBitstream->buffer_size,pFGSBitstream->pFGSFile);
  
  if(ferror(pFGSBitstream->pFGSFile))
  {
    snprintf(errortext, ET_SIZE, "initializeBitstream: The read operation has failed");
    error(errortext, 100);
  }
}


/*!
 ***********************************************************************
 * \brief
 *    closes the bitstream
 * \return
 *    none
 ***********************************************************************
 */

void closeBitstream()
{
  fclose(pFGSBitstream->pFGSFile);
  free(pFGSBitstream->streamBuffer);
  free(pFGSBitstream);
}

/*!
 ***********************************************************************
 * \brief
 *    reads the next numBits in the bitstream
 * \param data
 *    byte to read
 * \param pBitstream
 *    number of bits to read (0 to 8)
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    The read pointer is set accordingly
 * \return
 *    (+1) if the markers don't occur (-1) if they occur
 ***********************************************************************
 */

int readByteFGS(byte *data,unsigned int numBits, decBitstream *pBitstream)
{
  int bcounter,temp;
  if(numBits > 8)
  {
    printf("readByteFGS: numBits exceeds 8\n");
    exit(1);
  }
  if(numBits == 0) return 1;

  if((pBitstream->byte_pos +4) > pBitstream->read_bytes)
  {
    reallocBuffer(pBitstream);
  }

  bcounter = pBitstream->bit_pos - numBits;

  if(bcounter < 0)
    temp = pBitstream->bit_pos;
  else
    temp = numBits;

  *data = 0;
  *data = (pBitstream->streamBuffer[pBitstream->byte_pos] 
              >> (pBitstream->bit_pos-temp)) & mask_bits[temp-1];

  if(bcounter < 0)
  {
    pBitstream->byte_pos++;
    temp = ABS(bcounter);
    pBitstream->bit_pos = 8-temp;
    *data = ((pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos] 
              >> (8-temp)) & mask_bits[temp-1]) | (*data << temp);
    if(checkEnd(pBitstream) == -1) return -1;
  }
  else
  {
    pBitstream->bit_pos-=numBits;
  }
#ifdef FGS_TRACE
/*  if(gtr == 0)
  {
    fprintf(pFGSTraceFile,"Data:%x Lenght:%d\n",*data,numBits);
    fflush(pFGSTraceFile);
  }*/
#endif
  return 1;
}

/*!
 ***********************************************************************
 * \brief
 *    reads the next numBits in the bitstream
 * \param data
 *    int to read
 * \param pBitstream
 *    number of bits to read (0 to 32)
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    The read pointer is set accordingly
 * \return
 *    (+1) if the markers don't occur (-1) if they occur
 ***********************************************************************
 */
int readIntFGS(unsigned int *data, int numBits, decBitstream *pBitstream)
{
  int ret,div,remainder;
  byte temp;
  if(numBits > 32)
  {
    printf("readIntFGS: numBits exceeds 32\n");
    exit(1);
  }
  if(numBits == 0) return 1;

  div = (int)(numBits/8);
  remainder = numBits%8;
  *data = 0;
  
  switch(div)
  {
  case 4:
    ret = readByteFGS(&temp,8,pBitstream);
    *data = temp << (remainder+24);
    if(ret == -1) return ret;
  case 3:
    ret = readByteFGS(&temp,8,pBitstream);
    *data = *data | temp << (remainder+16);
    if(ret == -1) return ret;
  case 2:
    ret = readByteFGS(&temp,8,pBitstream);
    *data = *data | (temp << (remainder+8));
    if(ret == -1) return ret;
  case 1:
    ret = readByteFGS(&temp,8,pBitstream);
    *data = *data | (temp << remainder);
    if(ret == -1) return ret;
  case 0:
    ret = readByteFGS(&temp,remainder,pBitstream);
    *data = *data | temp;
    if(ret == -1) return ret;
  }
  return 1;
}

/*!
 ***********************************************************************
 * \brief
 *    Reads stuffing bytes and advances pointer
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    for byte aligned reads
 * \return
 *    none
 ***********************************************************************
 */
void readStuffingBits(decBitstream *pBitstream,int do_test)
{
  int numBits;
  byte temp,test;

  if(pBitstream->bit_pos == 0) 
    numBits = 8;
  else
    numBits = pBitstream->bit_pos;

  readByteFGS(&temp,numBits,pBitstream);
  pBitstream->byte_pos++;
  pBitstream->bit_pos = 8;

  if(do_test == 0)
  {
    test = (1<<(numBits-1)) -1;
    if(test != temp)
    {
      snprintf(errortext, ET_SIZE, "readStuffingBits: Wrong Code\n");
      error(errortext, 100);
    }
  }
}
/***********************************************************************
 * \brief
 *    shows the next numBits in the bitstream
 * \param data
 *    byte to read
 * \param pBitstream
 *    number of bits to read (0 to 8)
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    without advancing the read pointer
 * \return
 *    none
 ***********************************************************************
 */

void peekByteFGS(byte *data,unsigned int numBits, decBitstream *pBitstream)
{
  int bcounter,temp;
  if(numBits > 8)
  {
    printf("readByteFGS: numBits exceeds 8\n");
    exit(1);
  }
  if(numBits == 0) return;
  
  if((pBitstream->byte_pos +2) > pBitstream->read_bytes)
  {
    reallocBuffer(pBitstream);
  }

  bcounter = pBitstream->bit_pos - numBits;

  if(bcounter < 0)
    temp = pBitstream->bit_pos;
  else
    temp = numBits;

  *data = (pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos] 
              >> (pBitstream->bit_pos-temp)) & mask_bits[temp-1];

  if(bcounter < 0)
  {
    temp = ABS(bcounter);
    *data = ((pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos+1] 
              >> (8-temp)) & mask_bits[temp-1]) | (*data << temp);
  }
  return;
}

/*!
 ***********************************************************************
 * \brief
 *    shows the next numBits in the bitstream
 * \param data
 *    int to read
 * \param pBitstream
 *    number of bits to read (0 to 32)
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    without advancing the read pointer
 * \return
 *    none
 ***********************************************************************
 */

void peekIntFGS(unsigned int *data, int numBits, decBitstream *pBitstream)
{
  int div,remainder,bit_pos_temp,byte_pos_temp;
  byte temp;
  if(numBits > 32)
  {
    printf("peekIntFGS: numBits exceeds 32\n");
    exit(1);
  }
  if(numBits == 0) return;

  div = (int)(numBits/8);
  remainder = numBits%8;
  *data = 0;

  bit_pos_temp = pBitstream->bit_pos;
  byte_pos_temp = pBitstream->byte_pos;
  
  switch(div)
  {
  case 3:
    readByteFGS(&temp,8,pBitstream);
    *data = temp << (remainder+16);
  case 2:
    readByteFGS(&temp,8,pBitstream);
    *data = *data | (temp << (remainder+8));
  case 1:
    readByteFGS(&temp,8,pBitstream);
    *data = *data | (temp << remainder);   
  case 0:
    readByteFGS(&temp,remainder,pBitstream);
    *data = *data | temp;
  }
  pBitstream->bit_pos = bit_pos_temp;
  pBitstream->byte_pos = byte_pos_temp;
  return;
}

/*!
 ***********************************************************************
 * \brief
 *    Check if certain markers occur
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    Useful to decode a truncated bitstream
 * \return
 *    (+1) if the markers don't occur (-1) if they occur
 ***********************************************************************
 */

int checkEnd(decBitstream *pBitstream)
{
  int start_code;
  start_code = 0;
  start_code = (start_code<<8) | pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos];
  start_code = (start_code<<8) | pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos+1];
  start_code = (start_code<<8) | pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos+2];
  start_code = (start_code<<8) | pFGSBitstream->streamBuffer[pFGSBitstream->byte_pos+3];
  if(start_code == FGS_VOP_START_CODE || start_code == VISUAL_OBJECT_SEQUENCE_END_CODE)
  {
    pFGSData->endOfData = -1;
    return (-1);
  }
  else return 1;
}

/*!
 ***********************************************************************
 * \brief
 *    Bookmark a current position in the bitstream
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    Useful to peek byte aligned values
 * \return
 *    none
 ***********************************************************************
 */

void addBookmark(decBitstream *pBitstream)
{
  pBitstream->bit_pos_book = pBitstream->bit_pos;
}

/*!
 ***********************************************************************
 * \brief
 *    The current position in the bitstream is set to the bookmark value
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    Useful to peek byte aligned values
 * \return
 *    none
 ***********************************************************************
 */

void getBookmark(decBitstream *pBitstream)
{
  pBitstream->bit_pos = pBitstream->bit_pos_book;
}

/*!
 ************************************************************************
 * \brief
 *    Reallocates memory buffer and reads the next block of the file.
 *    Is called if all the content of the buffer is read, i.e. if an encoded 
 *    frame is larger than the size of the buffer (buffer_size)
 *    
 * \param pBitstream
 *    Bitstream buffer
 *
 ************************************************************************
 */
void reallocBuffer(decBitstream *pBitstream)
{
  int bytes_read;

  if((pBitstream->streamBuffer = realloc(pBitstream->streamBuffer, pBitstream->buffer_size 
    + pBitstream->read_bytes)) ==  NULL)
  {
    snprintf(errortext, ET_SIZE, "reallocBuffer: Memory allocation for FGS bitstream buffer failed");
    error(errortext, 100);
  }

  bytes_read = fread(pBitstream->streamBuffer+pBitstream->read_bytes,
    sizeof(byte),pBitstream->buffer_size,pBitstream->pFGSFile);

  if(ferror(pBitstream->pFGSFile))
  {
    snprintf(errortext, ET_SIZE, "reallocBuffer: The read operation has failed");
    error(errortext, 100);
  }
  pBitstream->read_bytes = pBitstream->read_bytes + bytes_read;
}

/*!
 ************************************************************************
 * \brief
 *    Discards all the data already read!!!
 *    Reads a buffer_size of data
 *    Assumes that the frame data starts at a byte aligned position
 *
 * \param pBitstream
 *    Bitstream buffer
 *
 * \param pData
 *    pointer to where the frame starts (relative to the buffer array)
 *   
 ************************************************************************
 */

void readNextFrame(decBitstream *pBitstream, int pdata)
{
  //shrink buffer to original value
  free(pBitstream->streamBuffer);
  pBitstream->streamBuffer = (byte *) calloc(pBitstream->buffer_size, sizeof(byte));
  if (pBitstream == NULL || pBitstream->streamBuffer == NULL)
  {
    snprintf(errortext, ET_SIZE, "readNextFrame: Memory allocation for FGS bitstream buffer failed");
    error(errortext, 100);
  }
  //position file pointer in the start of a new frame
  fseek(pBitstream->pFGSFile,pFGSBitstream->p_frame+pdata,SEEK_SET);
  pFGSBitstream->p_frame = pFGSBitstream->p_frame+pdata;

  //read next chunk of data
  pBitstream->read_bytes = fread(pBitstream->streamBuffer,sizeof(byte),
    pBitstream->buffer_size,pBitstream->pFGSFile);
  //init pointers
  pFGSBitstream->byte_pos = 0;
  pFGSBitstream->bit_pos = 8;
  
  if(ferror(pBitstream->pFGSFile))
  {
    snprintf(errortext, ET_SIZE, "readNextFrame: The read operation has failed");
    error(errortext, 100);
  }
}

void debug_residue(ImageParameters *pimg)
{
  int a,b;

  for(a=0; a<(pimg->height>>2); a++)
  {
    for(b=0; b<(pimg->width>>2); b++)
    {
      fread(&(pFGSData->Ydiff_dc[a][b]),sizeof(int),1,pFGSBitstream->pFGSFile);
    }
  }
  
  for(a=0; a<(pimg->height_cr>>2)*(pimg->width_cr>>2); a++)
  {
    fread(&(pFGSData->Udiff_dc[a]),sizeof(int),1,pFGSBitstream->pFGSFile);
  }
  
  for(a=0; a<(pimg->height_cr>>2)*(pimg->width_cr>>2); a++)
  { 
    fread(&(pFGSData->Vdiff_dc[a]),sizeof(int),1,pFGSBitstream->pFGSFile);
  }

  for(a=0; a<pimg->height; a++)
  {
    for(b=0; b<pimg->width; b++)
    {
      fread(&(pFGSData->Ydiff_ac[a][b]),sizeof(int),1,pFGSBitstream->pFGSFile);
    }
  }
  
  for(a=0; a<pimg->height_cr; a++)
  {
    for(b=0; b<pimg->width_cr; b++)
    {
      fread(&(pFGSData->Udiff_ac[a][b]),sizeof(int),1,pFGSBitstream->pFGSFile);
    }
  }
  
  for(a=0; a<pimg->height_cr; a++)
  {
    for(b=0; b<pimg->width_cr; b++)
    {
      fread(&(pFGSData->Vdiff_ac[a][b]),sizeof(int),1,pFGSBitstream->pFGSFile);
    }
  }
}


/*  i = 0;
  for(a=0; a<(4*pFGSData->hDCTMBlocks); a+=2)
  {
    for(b=0; b<(4*pFGSData->wDCTMBlocks); b+=2)
    {
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          pFGSData->YSignal_ac[a*TSIZE+c][b*TSIZE+d] = ldata_ac[i]->sign[0][fgs_zigzag[c*TSIZE+d]];
          pFGSData->YSignal_ac[a*TSIZE+c][(b+1)*TSIZE+d] = ldata_ac[i]->sign[1][fgs_zigzag[c*TSIZE+d]];
          pFGSData->YSignal_ac[(a+1)*TSIZE+c][b*TSIZE+d] = ldata_ac[i]->sign[2][fgs_zigzag[c*TSIZE+d]];
          pFGSData->YSignal_ac[(a+1)*TSIZE+c][(b+1)*TSIZE+d] = ldata_ac[i]->sign[3][fgs_zigzag[c*TSIZE+d]];
        }
      }
      i++;
    }
  }

  //get the sign of each AC chrominance and DC luminance coefficient in zig-zag scan order
  i = 0;
  for(a=0; a<(2*pFGSData->hDCTMBlocks); a+=2)
  {
    for(b=0; b<(2*pFGSData->wDCTMBlocks); b+=2)
    {
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          pFGSData->USignal_ac[a*TSIZE+c][b*TSIZE+d] = udata_ac[i]->sign[0][fgs_zigzag[c*TSIZE+d]];
          pFGSData->USignal_ac[a*TSIZE+c][(b+1)*TSIZE+d] = udata_ac[i]->sign[1][fgs_zigzag[c*TSIZE+d]];
          pFGSData->USignal_ac[(a+1)*TSIZE+c][b*TSIZE+d] = udata_ac[i]->sign[2][fgs_zigzag[c*TSIZE+d]];
          pFGSData->USignal_ac[(a+1)*TSIZE+c][(b+1)*TSIZE+d] = udata_ac[i]->sign[3][fgs_zigzag[c*TSIZE+d]];
          
          pFGSData->VSignal_ac[a*TSIZE+c][b*TSIZE+d] = vdata_ac[i]->sign[0][fgs_zigzag[c*TSIZE+d]];
          pFGSData->VSignal_ac[a*TSIZE+c][(b+1)*TSIZE+d] = vdata_ac[i]->sign[1][fgs_zigzag[c*TSIZE+d]];
          pFGSData->VSignal_ac[(a+1)*TSIZE+c][b*TSIZE+d] = vdata_ac[i]->sign[2][fgs_zigzag[c*TSIZE+d]];
          pFGSData->VSignal_ac[(a+1)*TSIZE+c][(b+1)*TSIZE+d] = vdata_ac[i]->sign[3][fgs_zigzag[c*TSIZE+d]];
        }
      }
      i++;
    }
  }*/