/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 ************************************************************************
 *  \file
 *     fgs_encoder.h
 *  \brief
 *     definitions for FGS encoding
 *  \author
 *      - Ascenso, Joo                             <joao.ascenso@lx.it.pt>
 *
 * ************************************************************************
 */

#ifndef FGS_ENCODER_H
#define FGS_ENCODER_H

/**************************************************************************
Definitions for the DCT Transform
***************************************************************************/

#define CONST_BITS  13
#define PASS1_BITS  2
#define FIX_0_298631336  ((int)  2446)
#define FIX_0_390180644  ((int)  3196)
#define FIX_0_541196100  ((int)  4433)
#define FIX_0_765366865  ((int)  6270)
#define FIX_0_899976223  ((int)  7373)
#define FIX_1_175875602  ((int)  9633)
#define FIX_1_501321110  ((int)  12299)
#define FIX_1_847759065  ((int)  15137)
#define FIX_1_961570560  ((int)  16069)
#define FIX_2_053119869  ((int)  16819)
#define FIX_2_562915447  ((int)  20995)
#define FIX_3_072711026  ((int)  25172)

#define MULTIPLY(var,const)  ((var) * (const))
#define RIGHT_SHIFT(x,shft)	((x) >> (shft))
#define DCTSIZE		    8
#define DCTSIZE2	    64
#define MBSIZE		    16
#define BSIZE		      8
#define TSIZE		      4
#define TSIZE2	      16
#define ONE	((int) 1)
#define DESCALE(x,n)  RIGHT_SHIFT((x) + (ONE << ((n)-1)), n)

/**************************************************************************
FGS definitions
***************************************************************************/

#define MAX_BP_LEVEL 15                 //maximum number of bitplanes supported
#define FGS_VOP_START_CODE 0x01B9       // use this one for now
#define FGS_VOP_START_CODE_LENGTH 32

#define FGS_VOP_BP_MARKER_HEADER 0x0A        // first 27 bits of fgs_vop_bp_marker
#define FGS_VOP_BP_MARKER_HEADER_LENGTH 27
#define FGS_VOP_BP_MARKER_ID_LENGTH 5

#define VISUAL_OBJECT_SEQUENCE_END_CODE 0x1B1
#define VISUAL_OBJECT_SEQUENCE_END_CODE_LENGHT 32

// to mask the n least significant bits of a byte or integer
static unsigned int mask_bits[9] =
{
    0x00000001, 0x00000003, 0x00000007,
    0x0000000f, 0x0000001f, 0x0000003f, 0x0000007f,
    0x000000ff
};

typedef struct enh_block      //!< enhancement block for the AC components
{
  int run[4][TSIZE2];				  //!< number of runs for each level
  int end[4][TSIZE2];				  //!< end = 0 or 1 for each level
  int sign[4][TSIZE2];        //!< sign of the coefficient in the block
  int data_level;             //!< number of bitplanes necessary to represent the block
  int cbp;                    //!< code bit pattern for each block
} enh_block;


typedef struct
{
  FILE *pFGSFile;              //!< pointer to the bitstream file
  FILE *pFGSTraceFile;         //!< pointer to trace file
  int **Ydiff_ac;              //!< residual images (AC components)
  int **Udiff_ac;
  int **Vdiff_ac;
  int **Ydiff_dc;              //!< residual images (DC components)
  int *Udiff_dc;
  int *Vdiff_dc;
  byte **YBitplane_ac;            //!< Y,U and V bitplane (AC components)
  byte **UBitplane_ac;
  byte **VBitplane_ac;

  byte **YBitplane_dc;            //!< Y,U and V bitplane (DC components)
  byte *UBitplane_dc;
  byte *VBitplane_dc;


  enh_block **block_data_lum_ac;  //!< pointer to vector of enhancenment blocks
  enh_block **block_data_u_ac;
  enh_block **block_data_v_ac;

  enh_block **block_data_lum_dc;  //!< pointer to vector of enhancenment blocks
  enh_block **block_data_u_dc;
  enh_block **block_data_v_dc;

  byte *except_code;           //!< exception code for fgs_msb_not_reached 

  int hDCTMBlocks;              //!< number of lum Mblocks in height
  int wDCTMBlocks;              //!< number of lum Mblocks in width

  int hDCTBlocks_cr;           //!< number of chr blocks in height
  int wDCTBlocks_cr;           //!< number of chr blocks in width

  byte     maximum_level_y_ac;    //!< maximum number of bitplanes of the Y AC component 
  byte     maximum_level_y_dc;    //!< maximum number of bitplanes of the Y DC component 

  byte     maximum_level_chr_dc;    //!< maximum number of bitplanes of the U&V DC component
  byte     maximum_level_chr_ac;    //!< maximum number of bitplanes of the U&V AC component
  byte     maximum_level;         //!< maximum number of bitplanes of all components

  Boolean skip_y_ac;                 //!< flag to indicate encoding of the y component
  Boolean skip_y_dc;                 //!< flag to indicate encoding of the y component

  Boolean skip_chr_dc;                 //!< flag to indicate encoding of the v component
  Boolean skip_chr_ac;                 //!< flag to indicate encoding of the u component

}FGSStruct;
/**************************************************************************
Useful macros
***************************************************************************/

#define SIGN0(a)    	(((a)<0) ? -1 : (((a)>0) ? 1  : 0))
#define SIGN(a)       ((a)<0 ? (-1) : (1))
#define ROUND(x)      ((int)((x) + SIGN0(x)*0.5))
#define ABS(x)        (((x) < 0) ? -(x) : (x))

/**************************************************************************
Prototype definitions           
***************************************************************************/

void initFGS (ImageParameters *pimg,InputParameters *pinp);
void cleanFGS();
void encodeFGSFrame (ImageParameters *pimg,InputParameters *pinp);
void createResidue(ImageParameters *pimg);

//transform related functions
void calcftrans_lum(int ** dataAC, int ** dataDC,int hDCTBlocks,int wDCTBlocks);
void calcftrans_chr(int ** dataAC, int * dataDC,int hDCTBlocks,int wDCTBlocks);
void H26L_transChr(int ** coefAC,int *coefDC);
void H26L_transLum(int **coefAC, int **coefDC);
void zigzag_scan(int hIndex,int wIndex,int i, byte **buffer,byte **bitplane);
void scanSignal(int hIndex,int wIndex,int sign[][TSIZE2],int **bitplane);

void findMaxLevel(ImageParameters *pimg);
int  findLevel(int mx);
void calcBitplane(ImageParameters *pimg,int level);
void outputFGSFrameHeader(Bitstream *pBitstream,int tref,int nBp);
void encodeBitplane(Bitstream *pBitstream,int level,int max_level);
void rleEncoding(byte *in, enh_block *block_data,int t_block);
void rleEncoding2x2(byte *in, enh_block* block_data, int t_block);
int getMBIndex(int level);
int getCbpIndex(int mbIndex,int blockIndex,int level);
void writeCbpindexMSB(int comb_index,int block_index,int cbp_index,byte *except_code);
void writeCbpindex(int comb_index,int block_index,int cbp_index,byte *except_code);
void initEncode();
void encodeBlockBP_lumDC(enh_block* block_data, int level);
void encodeBlockBP_chrDC(enh_block* block_data, int level, int max_level);
void encodeBlockBP(enh_block* block_data, int level, int max_level);
int nr_ones(int number);

// bitstream buffer management functions.. don't seem to find them on H.26L software
void writeByteFGS(byte data,unsigned int numBits, Bitstream *pBitstream);
void writeIntFGS(unsigned int data, int nbits, Bitstream *pBitstream);
void initializeBitstream(Bitstream *pBitstream);
void flushBuffer(Bitstream *pBitstream,FILE *file);
void insertStuffing(Bitstream *pBitstream);

//debug -> to delete
void debug_residue(ImageParameters *pimg);

#endif